from PyQt6.QtWidgets import QDialog, QLabel, QVBoxLayout, QFrame, QHBoxLayout

from otlmow_gui.Domain.network.Updater import Updater


class MenuActionsWindow:

    def __init__(self, language):
        self._ = language

    def create_about_window(self):
        window = QDialog()
        window_layout = QVBoxLayout()
        window.setMinimumWidth(500)
        window.setWindowTitle(self._('about'))
        version_box = self.version_box()
        window_layout.addWidget(version_box)
        mady_by_title = QLabel(self._('made_by') + ':')
        window_layout.addWidget(mady_by_title)
        window_layout.addWidget(self.creator_box('Jasper Berton', 'jasperberton1@telenet.be'))
        window_layout.addWidget(self.creator_box('David Vlaminck', 'david.vlaminck@mow.vlaanderen.be'))
        window_layout.addWidget(self.creator_box('Bert Van Overmeir','bert.vanovermeir@mow.vlaanderen.be'))
        window_layout.addWidget(
            self.creator_box('Christiaan Vanbergen', 'christiaan.vanbergen.btf@gmail.com'))
        window.setLayout(window_layout)
        window.exec()

    def create_error_report_window(self):
        window = QDialog()
        window_layout = QVBoxLayout()
        window.setMinimumWidth(500)
        window.setWindowTitle(self._('error_title'))
        error_title = QLabel(self._('error_report'))
        contact = QLabel(self._('contact') + ':')
        window_layout.addWidget(error_title)
        window_layout.addWidget(contact)
        window_layout.addWidget(self.creator_box('team BIM', 'TeamBim@verzendlijst.wegenenverkeer.be'))
        window.setLayout(window_layout)
        window.exec()

    def version_box(self):
        version_box = QFrame()
        version_box_layout = QHBoxLayout()
        version = Updater.get_project_version()

        version_title = QLabel(self._('version') + ':')
        version_number = QLabel(f'{version}')
        version_box_layout.addWidget(version_title)
        version_box_layout.addWidget(version_number)
        version_box.setLayout(version_box_layout)
        return version_box

    @classmethod
    def creator_box(cls, name, mail):
        creator_box = QFrame()
        creator_box_layout = QHBoxLayout()
        creator_name = QLabel(name)
        creator_box_layout.addWidget(creator_name)
        creator_mail = QLabel(f"<a href='mailto:{mail}'>{mail}</a>")
        creator_mail.setOpenExternalLinks(True)
        creator_box_layout.addWidget(creator_mail)
        creator_box.setLayout(creator_box_layout)
        return creator_box




import asyncio

from typing import List, Callable

from PyQt6.QtWidgets import QStackedWidget, QWidget

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.step_domain.ExportDataDomain import ExportDataDomain
from otlmow_gui.Domain.step_domain.ExportFilteredDataSubDomain import ExportFilteredDataSubDomain
from otlmow_gui.Domain.step_domain.HomeDomain import HomeDomain
from otlmow_gui.Domain.step_domain.InsertDataDomain import InsertDataDomain
from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.Domain.step_domain.TemplateDomain import TemplateDomain
from otlmow_gui.GUI.dialog_windows.NotificationWindow import NotificationWindow
from otlmow_gui.GUI.dialog_windows.YesOrNoNotificationWindow import YesOrNoNotificationWindow
from otlmow_gui.GUI.dialog_windows.file_picker_dialog.SubsetLoadFilePickerDialog import \
    SubsetLoadFilePickerDialog

from otlmow_gui.GUI.screens.ExportDataScreen import ExportDataScreen
from otlmow_gui.GUI.screens.HomeScreen import HomeScreen
from otlmow_gui.GUI.screens.InsertDataScreen import InsertDataScreen
from otlmow_gui.GUI.screens.RelationChangeScreen import RelationChangeScreen
from otlmow_gui.GUI.screens.Screen import Screen
from otlmow_gui.GUI.screens.TemplateScreen import TemplateScreen
from otlmow_gui.GUI.header.TabWidget import TabWidget
from otlmow_gui.GUI.screens.DynamicDataVisualisationScreen import DynamicDataVisualisationScreen


class MainWindow(QStackedWidget):
    def __init__(self, language: Callable):
        super().__init__()
        self._ = language

        self.home_screen:HomeScreen = HomeScreen(self._)
        self.step1:TemplateScreen = TemplateScreen(self._)
        self.step1_tabwidget:TabWidget = TabWidget(self._, page_nr=1, widget1=self.step1,
                                         description1="template",
                                         has_save_btn=False)
        self.step2:InsertDataScreen = InsertDataScreen(self._)
        self.step2_tabwidget:TabWidget = TabWidget(self._, page_nr=2, widget1=self.step2,
                                         description1="insert_data",
                                         has_save_btn=False)
        # self.step3_visuals:DataVisualisationScreen = DataVisualisationScreen(self._)
        self.step3_visuals: DynamicDataVisualisationScreen = DynamicDataVisualisationScreen(self._)
        # self.step3_map:MapScreen = MapScreen(self._)
        # self.step3_data:AssetDataChangeScreen = AssetDataChangeScreen(self._)
        self.step3_relations:RelationChangeScreen = RelationChangeScreen(self._)
        self.step_3_tabwidget:TabWidget = TabWidget(self._, page_nr=3, widget1=self.step3_relations,
                                          description1="relation_change",
                                          widget2=self.step3_visuals,
                                          description2="data visuals",
                                          # widget3=self.step3_map,
                                          # description3="map",
                                          has_save_btn=False)
        self.step4_export:ExportDataScreen = ExportDataScreen(self._)
        self.step4_tabwidget:TabWidget = TabWidget(self._, page_nr=4,
                                                widget1=self.step4_export,
                                                description1="export_data",
                                                has_save_btn=False)
        self.add_widget(self.home_screen)
        self.stepper_widgets:list[Screen] = [self.step1_tabwidget, self.step2_tabwidget, self.step_3_tabwidget,
                           self.step4_tabwidget]
        self.add_tabs_with_stepper_to_widget(self.stepper_widgets)
        self.home_screen.table.main_window = self
        self.step1.main_window = self

        HomeDomain.init_static(self.home_screen)
        self.subset_picker = SubsetLoadFilePickerDialog(self._)

        # dummy translation so the pybabel system doesn't remove them
        self._("template")
        self._("insert_data")
        self._("data visuals")
        self._("data_change")
        self._("relation_change")
        self._("export_data")

    def add_widget(self, widget: QWidget, has_stepper: bool = False):
        self.addWidget(widget)
        widget.main_window = self
        widget.header.main_window = self
        if has_stepper:
            widget.stepper_widget.main_window = self
            widget.tab1.main_window = self
            if hasattr(widget, 'tab2'):
                widget.tab2.main_window = self

    def reset_ui(self, _):
        for i in range(self.count()):
            self.widget(i).reset_ui(_)

    def add_tabs_with_stepper_to_widget(self, tabs: List[TabWidget]):
        for tab in tabs:
            self.add_widget(tab, True)

    def closeEvent(self, event):
        # Handle window close event
        OTLLogger.logger.debug("Checking if data visualisation is saved")
        if (global_vars.current_project and
            not global_vars.current_project.get_saved_graph_status()):

            title = self._("unsaved_graph_changes_warning_title")
            text = self._("unsaved_graph_changes_warning_text")
            alt_yes_text = self._("Continue")
            alt_no_text = self._("Back")

            warning_dialog = YesOrNoNotificationWindow(message=text,
                                                       title=title,
                                                       alt_yes_text=alt_yes_text,
                                                       alt_no_text=alt_no_text)
            answer = warning_dialog.exec()

            if answer == 16384:  # QMessageBox.ButtonRole.YesRole:
                # # save graph before closing application
                # global_vars.otl_wizard.main_window.step3_visuals.save_in_memory_changes_to_html()
                # asyncio.sleep(1)

                # close without doing anything else
                pass
            elif answer == 65536: # QMessageBox.ButtonRole.NoRole:
                # # close without doing anything else
                # pass

                # don't close return to application
                event.ignore()
                return
            elif answer == 262144: # QMessageBox.ButtonRole.AbortRole also X-button
                # don't close return to application
                event.ignore()
                return

        OTLLogger.logger.debug("Window is closing...")
        # Stop the asyncio event loop
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # Gather all pending tasks
            pending = asyncio.all_tasks(loop)
            for task in pending:
                task.cancel()  # Cancel all pending tasks

            loop.call_soon_threadsafe(loop.stop)


    def notify_user_of_excel_file_unavailable_error(self,e):
        message = self._(e.error_window_message_key)
        self.setWindowTitle(e.error_window_title_key)
        NotificationWindow("{0}:\n{1}".format(message, e.file_path))

    def go_to_project(self) -> None:
        self.setCurrentIndex(1)

    def enable_steps(self) -> None:
        self.reset_ui(self._)

    def setCurrentIndex(self, index):
        # if you go to the RelationChangeScreen or ExportDataScreen
        # the information is updated if the project has changed
        # if (index in [3, 4] and (not RelationChangeDomain.project or
        #                          RelationChangeDomain.project != global_vars.current_project)):
        # if (index in [3, 4] ):
        #     if RelationChangeDomain.cleared_data:
        #         RelationChangeDomain.init_static(project=global_vars.current_project)

        #everytime you go to a specific page update the frontend to always show the last Domain state
        if index == 1:
            TemplateDomain.update_frontend()
        elif index == 2:
            InsertDataDomain.update_frontend()
        elif index == 3:
            RelationChangeDomain.update_frontend()
            self.step3_visuals.opened()
        elif index == 4:
            ExportDataDomain.update_frontend()
            ExportFilteredDataSubDomain.update_frontend()

        super().setCurrentIndex(index)

    def update_color_scheme(self):
        # Import button on the home_screen
        self.home_screen.header.update_color_scheme()
        self.step3_relations.update_color_scheme()

        # the color of disabled steps in the step bar in a project
        self.step1_tabwidget.stepper_widget.enable_steps()
        self.step2_tabwidget.stepper_widget.enable_steps()
        self.step_3_tabwidget.stepper_widget.enable_steps()
        self.step4_tabwidget.stepper_widget.enable_steps()


    def show_blocking_yes_no_notification_window(self,text,title):
        msgbox = YesOrNoNotificationWindow(message=text,title=title)
        return msgbox.exec()


import asyncio
import datetime
import traceback
from pathlib import Path

from PyQt6.QtWidgets import QDialog, QLabel, QVBoxLayout
from PyQt6.QtGui import QFont, QMovie
from PyQt6.QtCore import Qt, QSize

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception

ROOT_DIR = Path(__file__).parent.parent.parent


def add_loading_screen(func):
    """Decorator that saves assets after executing the decorated function.

    This decorator wraps a function to ensure that after its execution, the current
    project's assets in memory are updated and saved. It also starts the event loop
    for the header in the main window to animate the OTL Wizard 2 logo during saving.

    :param func: The function to be decorated.
    :returns: The wrapper function that includes the saving logic.
    """

    async def wrapper_func(*args, **kwargs):
        if global_vars.test_mode:
            res = await func(*args, **kwargs)
            return res

        LoadingImageWindow.attempt_show_loading_screen(func.__name__)
        await asyncio.sleep(0)
        await asyncio.sleep(0)
        res = await func(*args, **kwargs)
        LoadingImageWindow.attempt_destoy_loading_screen(func.__name__)
        return res

    return wrapper_func

def add_loading_screen_no_delay(func):
    """Decorator that saves assets after executing the decorated function.

    This decorator wraps a function to ensure that after its execution, the current
    project's assets in memory are updated and saved. It also starts the event loop
    for the header in the main window to animate the OTL Wizard 2 logo during saving.

    :param delay: if there is a delay in opening the loading screen
    :param func: The function to be decorated.
    :returns: The wrapper function that includes the saving logic.
    """

    async def wrapper_func(*args, **kwargs):
        if global_vars.test_mode:
            res = await func(*args, **kwargs)
            return res

        LoadingImageWindow.attempt_show_loading_screen(func.__name__, delay=False)
        await asyncio.sleep(0)
        await asyncio.sleep(0)
        res = await func(*args, **kwargs)
        LoadingImageWindow.attempt_destoy_loading_screen(func.__name__)
        return res

    return wrapper_func


class LoadingImageWindow(QDialog):
    ref_key_to_time_dict = {}
    loading_window = None

    IMG_DIR = ProgramFileStructure.get_dynamic_library_path('img')

    image_path = IMG_DIR /  "cat_fly_animation.gif"
    # image_path = "C:\\Users\\chris\\PycharmProjects\\OTLMOW-GUI\\img\\yoda_patience.jpg"
    # image_path = "C:\\Users\\chris\\PycharmProjects\\OTLMOW-GUI\\img\\cat_fly_animation.gif"

    message = "Loading"
    title = "Loading"
    def __init__(self, parent=None, delayed_opening = False):
        super().__init__(parent)


        self.setWindowTitle(LoadingImageWindow.title)
        # self.setFixedSize(625, 468)
        self.setFixedSize(625, 350)
        self.setModal(True)  # Makes it a modal dialog
        self.setWindowFlag(Qt.WindowStaysOnTopHint)

        # Remove the close button
        # self.setWindowFlags(self.windowFlags() & ~Qt.WindowType.WindowCloseButtonHint)
        self.setWindowFlags(Qt.WindowType.CustomizeWindowHint | Qt.WindowType.WindowTitleHint)

        # Main layout
        layout = QVBoxLayout()

        # Image label
        image_abs_path_str = str(LoadingImageWindow.image_path.absolute())
        OTLLogger.logger.debug(f"loading_gif_path: {image_abs_path_str}")
        self.movie = QMovie(image_abs_path_str)
        self.movie.setScaledSize(QSize(625,468).scaled(int(625*1.5),int(468*1.5),Qt.AspectRatioMode.KeepAspectRatioByExpanding))

        movie_label = QLabel(self)
        movie_label.setMovie(self.movie)
        movie_label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        # Text label
        text_label = QLabel(LoadingImageWindow.message, self)
        text_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        text_label.setWordWrap(True)

        font = QFont()
        font.setPointSize(32)  # Adjust font size as needed
        text_label.setFont(font)



        # Add widgets to main layout
        layout.addWidget(movie_label)
        layout.addWidget(text_label)

        self.paintIntervals = []

        self.setLayout(layout)
        self.opening = True
        self.last_time = datetime.datetime.now()
        if delayed_opening:
            create_task_reraise_exception(self.delayed_open())
        else:
            self.open()
            self.movie.start()

    def paintEvent(self, *args, **kwargs):



        current_time = datetime.datetime.now()
        time_dif: datetime.timedelta = current_time - self.last_time
        self.last_time = current_time
        time_dif_seconds = time_dif.seconds + time_dif.microseconds / 1000000

        self.paintIntervals.append(time_dif_seconds)

    def closeEvent(self, event):
        # Ignore the close event to prevent closing with the 'X' button
        event.ignore()

    def close(self):
        self.opening = False

        if not self.isHidden():
            self.movie.stop()
            return super().close()


    async def delayed_open(self):
        # only open loading window after 1 second after initialisation
        # and only if the command to close it hasn't come yet by then
        OTLLogger.logger.debug("delayed open loadingscreen",
                               extra={"timing_ref": "open_loading_screen"})
        await asyncio.sleep(1)
        if self.opening:
            OTLLogger.logger.debug("still opening loadingscreen",
                                   extra={"timing_ref": "open_loading_screen"})
            self.open()
            self.movie.start()
            OTLLogger.logger.debug(f"movie scale {self.movie.scaledSize()}")
        else:
            OTLLogger.logger.debug("canceled opening loadingscreen",
                                   extra={"timing_ref": "open_loading_screen"})

    @classmethod
    def attempt_show_loading_screen(cls, ref: str,delay = True):
        # only create the loading screen if it hasn't already been created
        if not LoadingImageWindow.loading_window:
            loading_window = LoadingImageWindow(delayed_opening=delay)
            LoadingImageWindow.loading_window = {ref: loading_window}

    @classmethod
    def attempt_destoy_loading_screen(cls, ref: str):
        # only destory the loading screen in ref was the one who initiated the loading screen
        if LoadingImageWindow.loading_window:
            if ref in LoadingImageWindow.loading_window:

                current_loading_window = LoadingImageWindow.loading_window[ref]
                if current_loading_window.paintIntervals:
                    interval_count = len(current_loading_window.paintIntervals)
                    total_time = sum(current_loading_window.paintIntervals)
                    max_interval = max(current_loading_window.paintIntervals)
                    min_interval = min(current_loading_window.paintIntervals)
                    avg_interval =  total_time/interval_count
                    text = "Loading screen frame intervals: count: {interval_count}, tot: {tot:07.3f}s, max: {max:07.3f}s, min: {min:07.3f}s,avg: {avg:07.3f}s, all: [".format(interval_count=interval_count,tot=total_time,max=max_interval,min=min_interval,avg=avg_interval)
                    for time in current_loading_window .paintIntervals:
                        text += "{time:07.3f},".format(time=time)

                    text += "]"
                    OTLLogger.logger.debug(text)

                current_loading_window.close()
                LoadingImageWindow.loading_window = None
            elif ref == "crash":
                OTLLogger.logger.debug("ImageLoadingScreen closed by crash!!")
                stack = traceback.extract_stack(limit=3)
                OTLLogger.logger.debug(stack)
                current_loading_window = list(LoadingImageWindow.loading_window.values())[0]
                current_loading_window.close()
                LoadingImageWindow.loading_window = None



class RelationExistsForTargetButNotToTypeUriOfSource(Exception):
    """
    The bronAssetID.typeURI is INVALID for the current relationTypeURI and doelAssetID.typeURI
    For this error to occur the bronAssetID.typeURI or doelAssetID.typeURI need to exist first otherwise
    Error RelationHasNonExistingTypeUriForSourceOrTarget is thrown
    """
    def __init__(self,relation_type_uri,relation_identificator,wrong_field,wrong_value,*args):
        super().__init__(Exception(),*args)
        self.relation_type_uri = relation_type_uri
        self.relation_identificator = relation_identificator
        self.wrong_field = wrong_field
        self.wrong_value = wrong_value




from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QDialog, QLabel, QVBoxLayout, QHBoxLayout, QLineEdit, QDialogButtonBox, \
    QComboBox
from otlmow_model.OtlmowModel.Datatypes.KlBetrokkenheidRol import KlBetrokkenheidRol

from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain


class DefineHeeftBetrokkeneRelationWindow:
    options_to_data_dict = {}


    def __init__(self, language_settings,data_list_and_relation_objects):
        self.error_label = QLabel()
        self._ = language_settings

        self.input_asset_id_or_name: QLineEdit = QLineEdit()
        self.combobox_asset_type: QComboBox = QComboBox()

        self.data_list_and_relation_objects = data_list_and_relation_objects

    def draw_define_heeft_betrokkene_rol_window(self):

        self.error_label.setText("")
        dialog_window = QDialog()
        dialog_window.setMinimumWidth(450)
        # Makes the dialog the primary screen, disabling the screen behind it
        dialog_window.setModal(True)
        dialog_window.setWindowTitle(self._("needs_attribute"))
        # Creates the vertical stack layout
        layout = QVBoxLayout()

        # Creates 3 horizontal layouts for each input field with its label
        container_eigen_ref = QHBoxLayout()
        only_heeft_betrokkene_relation_objects =  [data for data in self.data_list_and_relation_objects
         if  data[2].typeURI == 'https://wegenenverkeer.data.vlaanderen.be/ns/onderdeel#HeeftBetrokkene']

        if(len(only_heeft_betrokkene_relation_objects) == 1):
            label_asset_id_or_name = QLabel( self._("define_attribute_to_add_heeft_betrokkene_relation").format(only_heeft_betrokkene_relation_objects[0][2].doelAssetId.identificator,only_heeft_betrokkene_relation_objects[0][2].bronAssetId.identificator))
            container_eigen_ref.addWidget(label_asset_id_or_name)
        else:
            label_asset_id_or_name = QLabel(
                self._("define_attribute_to_add_heeft_betrokkene_relations").format(
                    only_heeft_betrokkene_relation_objects[0][2].doelAssetId.identificator))
            container_eigen_ref.addWidget(label_asset_id_or_name)


        container_subset = QHBoxLayout()
        label_asset_type = QLabel(self._("asset_type") + ":")

        self.combobox_asset_type = self.create_combobox(KlBetrokkenheidRol.options)
        self.combobox_asset_type.setPlaceholderText(self._("asset_type_dummy"))
        container_subset.addWidget(label_asset_type, alignment=Qt.AlignmentFlag.AlignLeft)
        container_subset.addWidget(self.combobox_asset_type)

        # Adds the input fields to the layout
        layout.addLayout(container_eigen_ref)
        layout.addLayout(container_subset)

        # Changes the color of the error label to red
        self.error_label.setStyleSheet("color: red")

        # Creates the button box
        button_box:QDialogButtonBox = self.create_button_box()
        # sends the values off to validate once submitted
        button_box.accepted.connect(lambda: self.add_asset(dialog_window))
        button_box.rejected.connect(lambda: self.reject_action(dialog_window))
        # Adds the two buttons to the layout
        layout.addWidget(button_box)
        layout.addWidget(self.error_label)
        # Fills the dialog with the created layout
        dialog_window.setLayout(layout)
        # Shows the dialog
        dialog_window.show()
        dialog_window.exec()

    def add_asset(self, dialog_window):

        combobox_choice = self.combobox_asset_type.currentText()

        optie = KlBetrokkenheidRol.options[combobox_choice]

        RelationChangeDomain.last_added_to_existing.clear()
        for data in self.data_list_and_relation_objects:
            relation_object = data[2]

            if relation_object.typeURI == 'https://wegenenverkeer.data.vlaanderen.be/ns/onderdeel#HeeftBetrokkene':
                relation_object.rol = optie.invulwaarde

            RelationChangeDomain.add_relation_object_to_existing_relations(relation_object)
            RelationChangeDomain.last_added_to_existing.append(relation_object)

        RelationChangeDomain.update_frontend()
        dialog_window.close()

    def reject_action(self, dialog_window):

        RelationChangeDomain.update_frontend()
        dialog_window.close()

    def create_button_box(self):
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.setProperty("class", "button-box")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setProperty("class", "primary-button")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setText(self._("submit"))
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setProperty("class", "secondary-button")
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setText(self._("cancel"))
        return button_box

    def create_combobox(self, options_to_data_dict):
        comboBox = QComboBox()
        options = [key for key,option in options_to_data_dict.items() if option.status == "ingebruik" ]

        comboBox.addItems(options)
        # comboBox.currentTextChanged.connect(self.)

        return comboBox



from pathlib import Path

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.Domain.ProgramFileStructure import ProgramFileStructure
from otlmow_gui.Domain.database.SubsetDatabase import SubsetDatabase
from otlmow_gui.Domain.step_domain.ExportFilteredDataSubDomain import ExportFilteredDataSubDomain
from otlmow_gui.Domain.step_domain.InsertDataDomain import InsertDataDomain
from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain
from otlmow_gui.Domain.step_domain.TemplateDomain import TemplateDomain
from otlmow_gui.Exceptions.EmptyFieldError import EmptyFieldError
from otlmow_gui.Exceptions.NotASqlliteFileError import NotASqlliteFileError
from otlmow_gui.Exceptions.WrongDatabaseError import WrongDatabaseError
from otlmow_gui.GUI.dialog_windows.ProjectExistsError import ProjectExistsError
from otlmow_gui.GUI.screens.Screen import Screen
from otlmow_gui.GUI.screens.screen_interface.HomeScreenInterface import HomeScreenInterface

from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate


class HomeDomain:

    projects = {}

    @classmethod
    def init_static(cls, home_screen: HomeScreenInterface):
        cls.projects = cls.get_all_otl_wizard_projects()
        cls.update_frontend(screen=home_screen)

    @classmethod
    def get_all_otl_wizard_projects(cls) -> [Project]:
        """
        Retrieves all OTL wizard projects from the designated projects directory.

        This class method scans the directory for valid project directories,
        attempts to load each project, and collects them into a list,
        logging any directories that are not valid projects.

        :return: A list of loaded Project instances.
        :rtype: list[Project]

        :raises FileNotFoundError: If a project directory cannot be loaded due to missing files.
        """
        OTLLogger.logger.info(f"Execute HomeDomain.get_all_otl_wizard_projects",
                              extra={"timing_ref": f"get_all_otl_wizard_projects"})
        otl_wizard_project_dir = ProgramFileStructure.get_otl_wizard_projects_dir()

        project_dirs = [project_dir for project_dir in otl_wizard_project_dir.iterdir() if
                        project_dir.is_dir()]
        projects = {}
        for project_dir in project_dirs:
            try:
                project = Project.load_project(project_path=project_dir)
                projects[project.eigen_referentie] = project
            except FileNotFoundError:
                OTLLogger.logger.warning('Project dir %s is not a valid project directory', project_dir)

        project_count = len(projects)
        OTLLogger.logger.info(f"Execute HomeDomain.get_all_otl_wizard_projects ({project_count} projects)",
                              extra={"timing_ref": f"get_all_otl_wizard_projects"})

        return projects


    @classmethod
    def remove_project(cls, project: Project) -> None:
        """
        Removes a specified project and updates the project list in the frontend.

        This class method deletes the project's directory and refreshes the project data
        displayed to the user.

        :param project: The project instance to be removed.
        :type project: Project

        :return: None
        """

        project.delete_project_dir_by_path()
        cls.reload_projects()
        cls.update_frontend()

    @classmethod
    def validate(cls, input_eigen_ref: str, input_subset: str, db_path: str) -> bool:
        """
        Validates the input parameters for the operation. This class method checks that the
        provided eigen reference, subset, and database path are not empty and that the database
        path points to a valid subset database.

        The method raises specific errors if any of the input fields are empty or if the database
        path is invalid. If all checks pass, it returns True, indicating that the inputs are valid.

        :param input_eigen_ref: The eigen reference input to validate.
        :type input_eigen_ref: str
        :param input_subset: The subset input to validate.
        :type input_subset: str
        :param db_path: The path to the database to validate.
        :type db_path: str

        :return: True if all inputs are valid.
        :rtype: bool

        :raises EmptyFieldError: If any of the input fields are empty.
        :raises WrongDatabaseError: If the database path does not point to a valid subset database.
        """
        try:
            subset_db = SubsetDatabase(db_path=Path(db_path))
            is_valid_subset = subset_db.is_valid_subset_database()
            subset_db.close_connection()
        except NotASqlliteFileError:
            is_valid_subset = False

        if not input_eigen_ref.strip():
            raise EmptyFieldError(GlobalTranslate._('own_reference_empty_error'))
        elif not input_subset.strip():
            raise EmptyFieldError(GlobalTranslate._('bestek_empty_error'))
        elif not db_path.strip():
            raise EmptyFieldError(GlobalTranslate._('db_path_empty_error'))
        elif is_valid_subset is False:
            raise WrongDatabaseError(GlobalTranslate._('wrong_database_error'))
        else:
            return True

    @classmethod
    def change_subset(cls, new_path) -> None:
        """
        Changes the subset of a specified project and updates the associated UI components.
        This class method updates the project's last modified date, validates the new subset path,
         and refreshes the relevant information in the main window.

        The method first checks if the new subset path points to a valid database.
        If valid, it updates the project's subset and refreshes the UI to reflect the changes.

        :param project: The project instance whose subset is to be changed.
        :type project: Project
        :param new_path: The new path for the subset.
        :type new_path: str
        :param main_window: The main window instance to update the UI.
        :type main_window: QWidget

        :return: None

        :raises WrongDatabaseError: If the new path does not point to a valid subset database.
        """

        project = global_vars.current_project
        main_window = global_vars.otl_wizard.main_window

        if SubsetDatabase(db_path=Path(new_path)).is_valid_subset_database() is False:
            raise WrongDatabaseError("Wrong database")

        project.change_subset(new_path=Path(new_path))

        TemplateDomain.init_static()
        RelationChangeDomain.clear_data()
        RelationChangeDomain.init_static(project=project)

        main_window.reset_ui(GlobalTranslate._)

    @classmethod
    def get_screen(cls) -> HomeScreenInterface:
        return global_vars.otl_wizard.main_window.home_screen

    @classmethod
    def update_frontend(cls,screen: HomeScreenInterface = None)  -> None:
        """
        Updates the frontend display with the current list of projects.
        This class method fills the specified screen's table with project data, or
        defaults to the main screen if no screen is provided.

        The method checks if a specific screen is passed as an argument; if so, it updates that
        screen. Otherwise, it retrieves the main screen and updates it with the current projects.

        :param screen: The screen instance to update. If None, the main screen will be updated.
        :type screen: Screen or None

        :return: None
        """
        if screen:
            screen.fill_table(projects=cls.projects)
        else:
            cls.get_screen().fill_table(projects=cls.projects)

    @classmethod
    def open_project(cls, project_ref:str)  -> None:
        """
        Opens a specified project and prepares the application to work with it.

        This class method retrieves the project reference, sets it as the current project
        and loads any associated document filenames. It also updates the main window to reflect
        the selected project and initiates the process of filling the relevant UI components
        asynchronously.

        :param project_ref: The reference identifier for the project to be opened.
        :type project_ref: str

        :return: None
        """

        selected_project = cls.projects[project_ref]
        global_vars.current_project = selected_project

        global_vars.otl_wizard.main_window.go_to_project()

        selected_project.load_saved_document_filenames()

        TemplateDomain.init_static()
        InsertDataDomain.init_static()
        # RelationChangeDomain.clear_data()
        RelationChangeDomain.init_static(project=selected_project)
        ExportFilteredDataSubDomain.clear_data()
        global_vars.otl_wizard.main_window.step3_visuals.changed_project_bool = True

        global_vars.otl_wizard.main_window.enable_steps()



    @classmethod
    def process_upsert_dialog_input(cls, input_bestek: str, input_eigen_ref: str,
                                    input_subset: str, project: Project = None) -> None:
        """
        Processes input from the upsert dialog to create or update a project.

        This class method takes user input for project details and either creates a new project or
        updates an existing one based on the provided information. It then reloads the project list
        and updates the frontend to reflect the changes.

        :param input_bestek: The bestek input for the project.
        :type input_bestek: str
        :param input_eigen_ref: The eigen reference input for the project.
        :type input_eigen_ref: str
        :param input_subset: The subset input for the project.
        :type input_subset: str
        :param project: An optional Project instance to update. If None, a new project will be
        created.
        :type project: Project, optional

        :return: None
        """

        if project is None:
            project = cls.create_and_add_project(eigen_ref=input_eigen_ref,
                                       bestek=input_bestek,
                                       subset_path=Path(input_subset))
        else:
            project.update_information(new_eigen_ref=input_eigen_ref,
                                       new_bestek=input_bestek,
                                       new_subset_path=Path(input_subset))

        global_vars.otl_wizard.main_window.home_screen.last_added_ref = project.eigen_referentie
        cls.reload_projects()
        cls.update_frontend()

    @classmethod
    def create_and_add_project(cls, eigen_ref: str, bestek: str, subset_path: Path) -> Project:
        """
        Creates a new project and adds it to the project list.

        This class method initializes a Project instance with the provided eigen reference,
        bestek, and subset path, then saves the project to the designated directory. The newly
        created project is also added to the class's project collection for future reference.

        :param eigen_ref: The eigen reference for the new project.
        :type eigen_ref: str
        :param bestek: The bestek associated with the new project.
        :type bestek: str
        :param subset_path: The path to the subset for the new project.
        :type subset_path: Path

        :return: None
        """

        project = Project(eigen_referentie=eigen_ref, subset_path=subset_path, bestek=bestek)

        project_dir_path = project.get_project_local_path()
        if  project_dir_path.exists():
            OTLLogger.logger.error("Project dir %s already exists", project_dir_path)
            raise ProjectExistsError(eigen_referentie=project.eigen_referentie)

        cls.projects[eigen_ref] = project
        project.save_project_to_dir()

        return project

    @classmethod
    def reload_projects(cls) -> None:
        """
        Reloads the list of projects from the OTL wizard.

        This class method retrieves the current projects by calling a method that fetches all
        OTL wizard projects from disk storage and updates the class's project collection
        ccordingly.

        :return: None
        """

        cls.projects = cls.get_all_otl_wizard_projects()


from pathlib import Path

from PyQt6.QtWidgets import QTabWidget, QVBoxLayout

from otlmow_gui.Domain import global_vars
from otlmow_gui.GUI.dialog_windows.YesOrNoNotificationWindow import YesOrNoNotificationWindow
from otlmow_gui.GUI.screens.Screen import Screen
from otlmow_gui.GUI.header.HeaderBar import HeaderBar
from otlmow_gui.GUI.header.StepperWidget import StepperWidget

ROOT_DIR = Path(__file__).parent

LANG_DIR = ROOT_DIR.parent / 'locale/'


class TabWidget(Screen):

    def __init__(self, language, page_nr: int, widget1, description1: str, has_save_btn: bool = True, **kwargs):
        super().__init__()
        self._ = language
        self.tabs = QTabWidget()
        self.tab1 = widget1
        self.desc1 = description1
        self.stepper_widget = StepperWidget(self._, page_nr)
        self.header = HeaderBar(language=self._, has_save_btn=has_save_btn)
        self.main_window = None
        self.create_tab(widget1, **kwargs)
        self.layout = QVBoxLayout(self)
        self.fill_up_layout()

        self.init_ui()

    def create_tab(self, main_widget, **kwargs):
        self.tabs.addTab(main_widget, self._(self.desc1))
        widget2 = kwargs.get('widget2', None)
        widget3 = kwargs.get('widget3', None)
        if widget2 is not None:
            self.tab2 = widget2
            self.desc2 = kwargs.get('description2', None)
            self.tabs.addTab(self.tab2, self._(self.desc2))
            self.tabs.currentChanged.connect(self.tab2.opened)
        if widget3 is not None:
            self.tab3 = widget3
            self.desc3 = kwargs.get('description3', None)
            self.tabs.addTab(self.tab3, self._(self.desc3))

    def fill_up_layout(self):
        self.layout.addWidget(self.header)
        self.layout.addSpacing(10)
        self.layout.addWidget(self.stepper_widget.stepper_widget())
        self.layout.addSpacing(10)
        self.layout.addWidget(self.tabs)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.tabs.setProperty('class', 'tab-widget')

    def onClickReturnToHomeScreen(self):

        # first check if there are any unsaved changes in the graph
        if (global_vars.current_project and
            not global_vars.current_project.get_saved_graph_status()):

            title = self._("unsaved_graph_changes_warning_title")
            text = self._("unsaved_graph_changes_warning_text")
            alt_yes_text = self._("Continue")
            alt_no_text = self._("Back")

            warning_dialog = YesOrNoNotificationWindow(message=text,
                                                       title=title,
                                                       alt_yes_text=alt_yes_text,
                                                       alt_no_text=alt_no_text)
            answer = warning_dialog.exec()

            if answer == 16384:  # QMessageBox.ButtonRole.YesRole:
                # close project without doing anything else
                pass
            elif answer == 65536: # QMessageBox.ButtonRole.NoRole:
                # don't close project return to project
                return
            elif answer == 262144: # QMessageBox.ButtonRole.AbortRole also X-button
                # don't close project return to project
                return

        # Reset cached export names so new projects suggest their own names
        if (getattr(self.main_window, "step1", None) and
            getattr(self.main_window.step1, "template_export_file_picker", None)):
            self.main_window.step1.template_export_file_picker.previous_exported_file_name = None

        if (getattr(self.main_window, "step4_export", None) and
            getattr(self.main_window.step4_export, "sub_screen_option_1_all_data", None)):
            if hasattr(self.main_window.step4_export.sub_screen_option_1_all_data, "export_file_dialog_window"):
                self.main_window.step4_export.sub_screen_option_1_all_data.export_file_dialog_window.previous_exported_file_name = None

        if (getattr(self.main_window, "step4_export", None) and
            getattr(self.main_window.step4_export, "sub_screen_option_2_only_unedited_data", None)):
            if hasattr(self.main_window.step4_export.sub_screen_option_2_only_unedited_data, "export_file_dialog_window"):
                self.main_window.step4_export.sub_screen_option_2_only_unedited_data.export_file_dialog_window.previous_exported_file_name = None

        self.main_window.setCurrentIndex(0)
        global_vars.current_project.clear_model_builder_from_memory()
        global_vars.current_project = None
        self.main_window.home_screen.sort_on_last_edit()

    def init_ui(self):
        returnToHomeScreenButton = self.header.header_bar_detail_screen()
        returnToHomeScreenButton.clicked.connect(self.onClickReturnToHomeScreen)

    def reset_ui(self, _):
        super().reset_ui(_)
        self._ = _
        # OTLLogger.logger.debug("resetting a tab widget")
        self.header.reset_ui(_)
        if hasattr(self, 'tab2'):
            self.tab2.reset_ui(_)
            self.tabs.setTabText(1, self._(self.desc2))
        if hasattr(self, 'tab3'):
            self.tab3.reset_ui(_)
            self.tabs.setTabText(2, self._(self.desc3))
        # self.tab1.reset_ui(_)
        self.tabs.setTabText(0, self._(self.desc1))
        self.stepper_widget.reset_ui(_)



from PyQt6.QtWidgets import QTableView, QHeaderView


class ChangesTableView(QTableView):

    def __init__(self):
        super().__init__()
        self.did_first_paint = False
        self.did_second_paint = False
        self.previous_width = 0

    def paintEvent(self, e):
        super().paintEvent(e)

        curr_frame_width = self.frameRect().width()
        if not self.previous_width:
            self.previous_width = curr_frame_width
        elif curr_frame_width != self.previous_width:
            self.resetPainting()
            self.previous_width = curr_frame_width

        if not self.did_first_paint:
            self.did_first_paint = True
        elif not self.did_second_paint:
            hor_headers = self.horizontalHeader()
            hor_headers.setSectionResizeMode(QHeaderView.ResizeMode.Interactive)
            if hor_headers.count():
                self.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.ResizeToContents)
            self.did_second_paint = True


    def resetPainting(self):
        self.did_first_paint = False
        self.did_second_paint = False
        hor_headers = self.horizontalHeader()
        self.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
        if hor_headers.count():
            self.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.ResizeToContents)

    def setModel(self, model):
        super().setModel(model)
        self.resetPainting()





import asyncio
import gc
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import Optional, Union, Callable

from otlmow_converter.DotnotationDictConverter import DotnotationDictConverter
from otlmow_model.OtlmowModel.BaseClasses.OTLObject import dynamic_create_type_from_uri, \
    dynamic_create_instance_from_uri
from otlmow_model.OtlmowModel.BaseClasses.RelationInteractor import RelationInteractor
from otlmow_model.OtlmowModel.Classes.Agent import Agent
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.AIMObject import \
    AIMObject
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.RelatieObject import RelatieObject
from otlmow_model.OtlmowModel.Helpers import RelationCreator, OTLObjectHelper
from otlmow_model.OtlmowModel.Helpers.OTLObjectHelper import is_relation
from otlmow_model.OtlmowModel.Helpers.generated_lists import get_hardcoded_class_dict, \
    get_concrete_subclasses_from_class_dict
from otlmow_modelbuilder.OSLOCollector import OSLOCollector
from otlmow_modelbuilder.SQLDataClasses.OSLORelatie import OSLORelatie
from otlmow_gui.Domain import global_vars

from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.GUI.dialog_windows.LoadingImageWindow import add_loading_screen
from otlmow_gui.GUI.screens.RelationChange_elements.RelationChangeHelpers import RelationChangeHelpers
from otlmow_gui.GUI.screens.screen_interface.RelationChangeScreenInterface import \
    RelationChangeScreenInterface
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception

ROOT_DIR = Path(__file__).parent.parent

def async_save_assets(func):
    """Decorator that saves assets after executing the decorated function.

    This decorator wraps a function to ensure that after its execution, the current
    project's assets in memory are updated and saved. It also starts the event loop
    for the header in the main window to animate the OTL Wizard 2 logo during saving.

    :param func: The function to be decorated.
    :returns: The wrapper function that includes the saving logic.
    """

    async def wrapper_func(*args, **kwargs):
        res = await func(*args, **kwargs)
        global_vars.current_project.assets_in_memory = RelationChangeDomain.get_quicksave_instances()
        await global_vars.current_project.save_validated_assets()
        global_vars.otl_wizard.main_window.step_3_tabwidget.header.start_event_loop()
        return res

    return wrapper_func

class RelationChangeDomain:
    """
    RelationChangeDomain manages the relationships between various objects within a project. It provides methods for initializing project data, managing internal and external objects, and handling relations, including adding, removing, and retrieving relations.

    Attributes:
        project (Project): The current project being managed.
        collector (OSLOCollector): The collector for OSLO data.
        internal_objects (list[Union[RelatieObject, RelationInteractor]]): Objects placed by the contractor within the project.
        external_objects (list[Union[RelatieObject, RelationInteractor]]): Object outside the project (from DAVIE)
        agent_objects (list[Agent]): Agent objects that are not AIM objects.
        shown_objects (list[RelationInteractor]): All objects displayed in the GUI.
        possible_relations_per_class_dict (dict[str, list[OSLORelatie]]): Possible relations
            categorized by class type
        possible_object_to_object_relations_dict (dict[str, dict[str, list[RelatieObject]]]):
            Possible Relations between objects that the user can choose from.
        existing_relations (list[RelatieObject]): Currently active relations.
        aim_id_relations (list): Pre-existing relations sourced from DAVIE.
        selected_object (Optional[RelationInteractor]): The currently selected object from
            the OTL-asset column in the GUI.
        last_added_to_existing (Optional[list[RelatieObject]]): The last relations added to
            existing relations.
        last_added_to_possible (Optional[list[RelatieObject]]): The last relations added to
            possible relations.
        external_object_added (bool): Flag indicating if an external object has been added.

    Methods:
        init_static(project: Project) -> None: Initializes static resources for the class.
        load_project_relation_data() -> None: Loads project relation data asynchronously.
        set_instances(objects_list: list[Union[RelatieObject, RelationInteractor]]) -> None: Processes and categorizes AIM objects.
        create_and_add_missing_external_assets_from_relations() -> None: Creates missing external
            assets from relations.
        get_all_relations() -> list[RelatieObject]: Retrieves all existing relations.
        get_object(identificator: str) -> Optional[RelationInteractor]: Retrieves an object by its
            identifier.
        filter_on_id(id_to_check: str): Filters objects based on a specified identifier.
        set_possible_relations(selected_object: RelationInteractor) -> None: Sets possible
            relations for a selected object.
        get_possible_relations_for(selected_id: str) -> dict: Retrieves possible relations for a
            given ID.
        add_all_possible_relations_between_selected_and_related_objects(
            relation_list: list[OSLORelatie], selected_object: RelationInteractor,
            related_objects: list[RelationInteractor]) -> None: Adds possible relations between
            selected and related objects.
        add_inactive_relations_to_possible_relations(selected_id: str) -> None: Adds inactive
            relations to possible relations.
        are_possible_relations_to_other_class_types_collected_for(typeURI: str) -> bool: Checks if
            possible relations for a class type URI have been collected.
        collect_possible_relations_to_class_types_from(selected_object: RelationInteractor)
            -> None: Collects possible relations for a selected object.
        get_all_concrete_relation_from_full_model(selected_object: RelationInteractor)
            -> list[OSLORelatie]: Retrieves all concrete relations for a selected object.
        get_same_relations_in_list(relation_list: list[RelatieObject], relation_def: OSLORelatie,
            selected_object: RelationInteractor, related_object: RelationInteractor,
            reverse: bool = False) -> list: Retrieves relations matching a specified definition.
        is_same_relation(existing_relation: RelatieObject, relation_def: OSLORelatie,
            selected: RelationInteractor, related: RelationInteractor, reverse: bool = False)
            -> bool: Determines if an existing relation matches a specified definition.
        add_relation_between(relation: OSLORelatie, selected_object: RelationInteractor,
            related_object: RelationInteractor, reverse: bool = False) -> None: Adds a relation
            between two objects.
        add_possible_relation_object(relation_object, selected_object_id, related_object_id)
            -> None: Adds a relation object to possible relations.
        filter_out(object_to_filter_for: RelationInteractor) -> Callable: Creates a filter function
            to exclude a specified object.
        get_screen() -> RelationChangeScreenInterface: Retrieves the current screen interface.
        create_relation_object(OSLO_relation: OSLORelatie, source_object: RelationInteractor,
            target_object: RelationInteractor) -> Optional[RelatieObject]: Creates a relation
            object based on provided definitions.
        add_multiple_possible_relations_to_existing_relations(data_list): Adds multiple possible
            relations to existing relations.
        add_possible_relation_to_existing_relations( bron_asset_id:str, target_asset_id:str,
            relation_object_index:int) -> RelatieObject: Adds a possible relation to existing
            relations.
        add_relation_object_to_existing_relations(relation_object: RelatieObject): Adds a relation
            object to existing relations.
        update_frontend(): Updates the user interface to reflect the current state of objects and
            relations.
        remove_multiple_existing_relations(indices: list[int]) -> None: Removes multiple existing
            relations based on indices.
        remove_existing_relation(index: int) -> RelatieObject: Removes an existing relation by
            index.
        select_existing_relation_indices(indices: list[int]) -> None: Selects existing relations
            based on indices.
        select_possible_relation_data(selected_relations_data: list) -> None: Selects and displays
            data for possible relations.
        get_internal_objects() -> list[RelationInteractor]: Retrieves the list of internal objects.
        get_persistent_relations() -> list[RelatieObject]: Retrieves a combined list of persistent
            relations.
        get_export_instances() -> list[Union[RelatieObject, RelationInteractor]]: Retrieves a
            combined list of instances for export.
        get_quicksave_instances() -> list[Union[RelatieObject, RelationInteractor]]: Retrieves a
            combined list of instances for quicksave.
        apply_active_aim_id_relations() -> None: Applies active AIM ID relations to existing
            relations.
        get_inactive_aim_id_relations() -> list: Retrieves a list of inactive AIM ID relations.
        add_external_objects_to_shown_objects() -> None: Adds external objects to shown objects.
        add_agent_objects_to_shown_objects() -> None: Adds agent objects to shown objects.
        create_and_add_new_external_asset(id_or_name: str, type_uri: str) -> None: Creates and
            adds a new external asset.
    """

    project: Project = None
    collector: OSLOCollector = None

    internal_objects: list[RelationInteractor] = []  # Object in the project (placed by contractor)
    external_objects: list[RelationInteractor] = []  # Object outside the project (from DAVIE)
    agent_objects: list[Agent] = []  # Agent objects (not an AIMobject)
    shown_objects: list[RelationInteractor] = []  # All objects combined that are displayed on GUI (col 1)

    possible_relations_per_class_dict: dict[str, list[OSLORelatie]] = {}
    possible_object_to_object_relations_dict: dict[
        str, dict[str, list[RelatieObject]]] = {}  # (col 2)

    existing_relations: list[RelatieObject] = []  # relations
    aim_id_relations = []  # pre-existing relations from DAVIE

    selected_object: Optional[RelationInteractor] = None  # the asset/agent in col 1 selected by user
    last_added_to_existing: Optional[list[RelatieObject]] = []  # relation last added to col 3
    last_added_to_possible: Optional[list[RelatieObject]] = []  # relation last added to col 2

    regenerate_relation_types = False

    map_uptodate = True

    search_full_OTL_mode = False

    cleared_data = True

    @classmethod
    def init_static(cls, project: Project, asynchronous=True) -> None:
        """
        Initializes static resources for the RelationChangeDomain class.
        Call this when the project or project.subset_path changes or everytime you go to the window

        This method sets up the project and initializes various attributes related
        to object relations, including collectors and lists for managing internal
        and external objects. It also starts loading project relation data asynchronously.

        :param cls: The class itself.
        :param project: The project to be initialized.
        :type project: Project
        :returns: None
        """

        cls.project = project
        cls.collector = OSLOCollector(project.subset_path)
        cls.collector.collect_all()

        cls.clear_data()

        if not Helpers.all_OTL_asset_types_dict:
            Helpers.create_external_typeURI_options()

        if global_vars.current_project:
            if asynchronous:
                try:
                    create_task_reraise_exception(cls.load_project_relation_data_async())
                except DeprecationWarning:
                    # should only go here if you are testing
                    create_task_reraise_exception(cls.load_project_relation_data_async())
            else:
                try:
                    cls.load_project_relation_data()
                except DeprecationWarning:
                    # should only go here if you are testing
                    cls.load_project_relation_data()


    @classmethod
    def get_empty_visualisation_uptodate(cls):
        return {"remove":[],"add":[],"clear_all": False}


    @classmethod
    def clear_data(cls):
        cls.selected_object = None

        cls.shown_objects = []
        cls.internal_objects = []
        cls.external_objects = []
        cls.existing_relations = []
        cls.possible_relations_per_class_dict = {}
        cls.possible_object_to_object_relations_dict = {}
        cls.aim_id_relations = []
        cls.regenerate_relation_types = False
        global_vars.current_project.visualisation_uptodate.reset_relations_uptodate()
        # global_vars.current_project.visualisation_uptodate.set_clear_all(True)
        cls.map_uptodate = False
        cls.cleared_data = True

    @classmethod
    @add_loading_screen
    async def load_project_relation_data_async(cls) -> None:
        """
        Loads project relation data asynchronously.

        This method sets the GUI to a loading state while it retrieves and processes
        project relation data. It populates a dictionary with asset type URIs and
        updates the user interface accordingly once the data is loaded.

        :param cls: The class itself.
        :returns: None
        """
        timing_ref = f"load_project_data_async"
        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.load_project_relation_data_async() for project "
            f"{global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": timing_ref})
        cls.get_screen().set_gui_lists_to_loading_state()


        # throw away old data before loading the new
        cls.clear_data()
        gc.collect()

        await cls.set_instances(objects_list=await cls.project.load_validated_assets_async())
        # global_vars.otl_wizard.main_window.step3_visuals.reload_html()
        timing_ref = f"load_project_data_async"
        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.load_project_relation_data_async() for project "
            f"{global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": timing_ref})

    @classmethod
    async def load_project_relation_data(cls) -> None:
        """
        Loads project relation data asynchronously.

        This method sets the GUI to a loading state while it retrieves and processes
        project relation data. It populates a dictionary with asset type URIs and
        updates the user interface accordingly once the data is loaded.

        :param cls: The class itself.
        :returns: None
        """
        timing_ref = f"load_project_data"
        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.load_project_relation_data() for project "
            f"{global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": timing_ref})
        cls.get_screen().set_gui_lists_to_loading_state()

        # throw away old data before loading the new
        cls.clear_data()
        gc.collect()

        await cls.set_instances(objects_list=cls.project.load_validated_assets())
        # global_vars.otl_wizard.main_window.step3_visuals.reload_html()
        timing_ref = f"load_project_data"
        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.load_project_relation_data() for project "
            f"{global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": timing_ref})

    @classmethod
    async def set_instances(cls, objects_list: list[Union[RelatieObject, RelationInteractor]]) -> None:
        # sourcery skip: use-contextlib-suppress
        """Processes and categorizes AIM objects for relation management.

        This method takes a list of AIM objects and categorizes them into various
        groups based on their properties.
        It updates the class attributes to reflect the current state of these objects
        and ensures that the user interface is populated with the correct information.

        :param cls: The class itself.
        :param objects_list: A list of AIM objects to be processed and categorized.
        :type objects_list: list[Union[RelatieObject, RelationInteractor]]
        :returns: None
        """
        timing_ref = f"set_instances"
        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.set_instances() for project "
            f"{global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": timing_ref})

        cls.existing_relations = []
        cls.aim_id_relations = []

        cls.internal_objects = []
        cls.external_objects = []
        cls.agent_objects = []
        cls.shown_objects = []

        for instance in objects_list:
            await asyncio.sleep(0)
            if is_relation(instance):
                # noinspection PyTypeChecker
                relation_instance: RelatieObject = instance
                is_aim_id = False
                # noinspection PyBroadException
                try:
                    is_aim_id = OTLObjectHelper.is_aim_id(aim_id=relation_instance.assetId.identificator)
                except Exception:
                    pass

                if is_aim_id or not relation_instance.isActief:
                    cls.aim_id_relations.append(relation_instance)
                else:
                    cls.existing_relations.append(relation_instance)

            else:
                if instance.typeURI == 'http://purl.org/dc/terms/Agent':
                    # noinspection PyTypeChecker
                    agent_instance: Agent = instance
                    cls.agent_objects.append(agent_instance)
                elif instance.assetId.toegekendDoor == global_vars.external_toegekendDoor_label:
                    cls.external_objects.append(instance)
                else:
                    cls.internal_objects.append(instance)

        cls.apply_active_aim_id_relations()

        cls.shown_objects = deepcopy(cls.internal_objects)
        cls.add_external_objects_to_shown_objects()
        cls.add_agent_objects_to_shown_objects()

        await cls.create_and_add_missing_external_assets_from_relations()

        cls.get_screen().fill_object_list(cls.shown_objects)
        cls.get_screen().fill_possible_relations_list(None, {})
        cls.get_screen().fill_existing_relations_list(cls.existing_relations)
        global_vars.otl_wizard.main_window.step3_visuals.update_only_legend()
        global_vars.otl_wizard.main_window.step3_visuals.update_slider_range()

        cls.cleared_data = False
        cls.map_uptodate = False
        timing_ref = f"set_instances"
        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.set_instances() for project "
            f"{global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": timing_ref})

    @classmethod
    async def create_and_add_missing_external_assets_from_relations(cls) -> None:
        """
        Creates and adds missing external assets based on existing relations.

        This method iterates through all relations and checks for the existence of
        source and target objects. If either object is missing, it attempts to create
        and add a new external asset using the relation's identifiers and type URIs.

        :param cls: The class itself.
        :returns: None
        """
        timing_ref = f"create_missing_external_assets"
        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.create_and_add_missing_external_assets_from_relations() for project "
            f"{global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": timing_ref})


        for relation_object in RelationChangeDomain.get_all_relations():
            await asyncio.sleep(0)
            source_id =  relation_object.bronAssetId.identificator
            target_id = relation_object.doelAssetId.identificator

            source_object = RelationChangeDomain.get_object(identificator=source_id)
            id_to_typeURI_dict: Optional[dict[str, str]] = None
            if not source_object:
                try:
                    bron_type_uri = Helpers.extract_corrected_relation_partner_typeURI(
                        relation_object.bron.typeURI,
                        relation_object.bronAssetId.identificator,
                        id_to_typeURI_dict,
                        RelationChangeDomain.get_shown_objects())

                    cls.create_and_add_new_external_asset(id_or_name=source_id,
                                                          type_uri=bron_type_uri)
                except ValueError as e:
                    # should there be a wrong typeURI
                    OTLLogger.logger.debug(e)

            target_object = RelationChangeDomain.get_object(identificator=target_id)
            if not target_object:
                try:
                    doel_type_uri = Helpers.extract_corrected_relation_partner_typeURI(
                        relation_object.doel.typeURI,
                        relation_object.doelAssetId.identificator,
                        id_to_typeURI_dict,
                        RelationChangeDomain.get_shown_objects())

                    cls.create_and_add_new_external_asset(id_or_name=target_id,
                                                          type_uri=doel_type_uri)
                except ValueError as e:
                    # should there be a wrong typeURI
                    OTLLogger.logger.debug(e)

        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.create_and_add_missing_external_assets_from_relations() for project "
            f"{global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": timing_ref})

    @classmethod
    def get_all_relations(cls) -> list[RelatieObject]:
        """Retrieves all existing relations from the class.

        This method combines and returns the lists of existing relations and AIM ID
        relations stored in the class. It provides a unified view of all relations
        for further processing or display.

        :param cls: The class itself.
        :returns: A list containing all existing and AIM ID relations.
        :rtype: list
        """

        return cls.existing_relations + cls.aim_id_relations

    @classmethod
    def get_object(cls, identificator: str) -> Optional[RelationInteractor]:
        # sourcery skip: use-named-expression
        """
        Retrieves an object from the list of shown objects based on its identificator.

        This method filters the list of shown objects to find an object that matches
        the provided identificator. If multiple objects are found, it currently does
        not handle the error case, but if exactly one object is found, it returns that object.

        :param cls: The class itself.
        :param identificator: The identificator of the object to retrieve.
        :type identificator: str
        :returns: The matching OTL object if found, otherwise None.
        :rtype: Optional[RelationInteractor]
        """
        filtered_objects = list(filter(cls.filter_on_id(id_to_check=identificator), cls.shown_objects))

        if filtered_objects:
            if len(filtered_objects) > 1:
                OTLLogger.logger.debug(f"ERROR RelationChangeDomain.get_object function found multiple "
                              f"objects with id: {identificator}")
                pass  # TODO: raise error of 2 objects with the same identificator in the list
            else:
                return filtered_objects[0]
        return None

    @classmethod
    def filter_on_id(cls, id_to_check: str):
        """
        Filters objects based on a specified identifier.
        This class method returns a lambda function that checks if the identifier of a
        given object matches the provided ID.

        :param id_to_check: The identifier to compare against.
        :type id_to_check: str

        :return: A lambda function that takes an object and returns True if its identifier matches the specified ID, otherwise False.
        :rtype: function
        """

        return lambda otl_object: RelationChangeHelpers.get_corrected_identificator(
            otl_object) == id_to_check

    @classmethod
    async def set_possible_relations(cls, selected_object: RelationInteractor):
        """
        Sets the possible relations for the specified selected object and updates the user
        interface accordingly. This method manages the collection and display of relations based
        on the selected object's typeURI.

        :param selected_object: The object for which possible relations are to be displayed.
        :type selected_object: AIMObject

        :return: None

        :raises ValueError: If the selected_object is invalid or cannot be processed.

        :Examples:
            RelationChangeDomain.set_possible_relations(selected_object)
        """



        cls.set_selected_object(selected_object)
        if not cls.selected_object:
            cls.get_screen().clear_possible_relation_elements()
            return

        # generate all relation types that the selected asset can have (within subset or full OTL)
        # only update this list when an external asset is added
        # OR when the user switches to the full OTL
        if (cls.regenerate_relation_types or not
                RelationChangeDomain.are_possible_relations_to_other_class_types_collected_for(
                selected_object.typeURI)):
            cls.collect_possible_relations_to_class_types_from(selected_object)
            cls.regenerate_relation_types = False

        # get all assets except the selected_object
        candidate_partner_assets: list[RelationInteractor] = list(
            filter(RelationChangeDomain.filter_out(object_to_filter_for=selected_object), 
                   cls.shown_objects))

        selected_id = RelationChangeHelpers.get_corrected_identificator(otl_object=selected_object)
        relation_types_list = cls.possible_relations_per_class_dict[selected_object.typeURI]

        # remove all previously generated relations for the selected_id
        cls.possible_object_to_object_relations_dict[selected_id] = {}
        # add all relations loaded that have isActief == False
        cls.add_inactive_relations_to_possible_relations(selected_id=selected_id)

        cls.add_all_possible_relations_between_selected_and_related_objects(
            relation_list=relation_types_list,selected_object=selected_object,
            related_objects=candidate_partner_assets)

        # sort generated relations
        cls.possible_object_to_object_relations_dict = (
            Helpers.sort_nested_dict(dictionary=cls.possible_object_to_object_relations_dict))

        possible_relations_for_this_object = cls.get_possible_relations_for(selected_id=selected_id)

        # noinspection PyTypeChecker
        object_attributes_dict = await DotnotationDictConverter.to_dict_async(otl_object=selected_object)

        cls.get_screen().fill_possible_relations_list(source_object=selected_object,
                                                      relations=possible_relations_for_this_object,
                                                      last_added=cls.last_added_to_possible)
        cls.get_screen().fill_object_attribute_field(object_attribute_dict=object_attributes_dict)


    @classmethod
    def get_possible_relations_for(cls, selected_id=str):
        """
        Retrieves the possible relations for a given selected ID. This method checks if the selected ID exists in the internal dictionary of possible relations and returns the corresponding relations if found.

        :param selected_id: The identifier for which possible relations are to be retrieved.
        :type selected_id: Any

        :return: A dictionary of possible relations for the selected ID, or an empty dictionary if none are found.
        :rtype: dict
        """

        if selected_id in cls.possible_object_to_object_relations_dict:
            return cls.possible_object_to_object_relations_dict[selected_id]
        return {}

    @classmethod
    def set_selected_object(cls,selected_object:RelationInteractor) -> None:
        cls.selected_object = selected_object

    @classmethod
    def add_all_possible_relations_between_selected_and_related_objects(
            cls, relation_list: list[OSLORelatie],selected_object: RelationInteractor,
            related_objects: list[RelationInteractor]) -> None:
        """
        Adds all possible relations between a selected object and its related objects based on the
        provided relation list. This method checks existing relations to avoid duplicates and
        establishes new relations as necessary.

        :param relation_list: A list of relations that define how objects can be related.
        :type relation_list:  list[OSLORelatie]

        :param selected_object: The object for which relations are being established.
        :type selected_object: RelationInteractor

        :param related_objects: A list of objects that may be related to the selected object.
        :type related_objects:  list[RelationInteractor]

        :return: None
        """
        log_typeURI = RelationChangeHelpers.get_abbreviated_typeURI(selected_object.typeURI)
        selected_object_id: str = RelationChangeHelpers.get_corrected_identificator(
            selected_object)

        related_objects_dict = defaultdict(list)
        legacy_hoortbij_relations_dict = {}
        for related_object in related_objects:
            related_objects_dict[related_object.typeURI].append(related_object)
            # if the related object is a legacy object create a new OSLO relation from legacy to otl_class
            if related_object.typeURI.startswith("https://lgc"):
                hoortbij_relation = RelationChangeDomain.create_hoortbij_OSLORelatie_with_legacy_class(
                    legacy_class_typeURI=related_object.typeURI,
                    OTL_class_typeURI= selected_object.typeURI)
                legacy_hoortbij_relations_dict[related_object.typeURI] = hoortbij_relation

        # add the hoortbij OSLORelations to legacy classes to the list of possible relations
        relation_list.extend(list(legacy_hoortbij_relations_dict.values()))

        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.add_all_possible_relations_between_selected_and_related_objects({log_typeURI}) for project {global_vars.current_project.eigen_referentie}",
            extra={"timing_ref": f"add_possible_relations_real_objects"})
        for relation in relation_list:
            if relation.bron_uri == selected_object.typeURI:
                for related_object in related_objects_dict[relation.doel_uri]:
                    if cls.get_same_relations_in_list(
                            relation_list=cls.existing_relations, relation_def=relation,
                            selected_object=selected_object, related_object=related_object):
                        continue
                    cls.add_relation_between(relation=relation,selected_object=selected_object,
                                             related_object=related_object)

            if relation.doel_uri == selected_object.typeURI:
                for related_object in related_objects_dict[relation.bron_uri]:
                    if cls.get_same_relations_in_list(
                            relation_list=cls.existing_relations, relation_def=relation,
                            selected_object=selected_object, related_object=related_object,
                            reverse=True):
                        continue
                    cls.add_relation_between(relation=relation,selected_object=selected_object,
                                             related_object=related_object, reverse=True)

        relation_count = 0
        if (cls.possible_object_to_object_relations_dict[selected_object_id]):
            for rel_obj in cls.possible_object_to_object_relations_dict[selected_object_id].keys():
                relation_count += len(cls.possible_object_to_object_relations_dict[selected_object_id][rel_obj])

        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.add_all_possible_relations_between_selected_and_related_objects({log_typeURI}) for project {global_vars.current_project.eigen_referentie} ({relation_count} relations)",
            extra={"timing_ref": f"add_possible_relations_real_objects"})

    @classmethod
    def add_inactive_relations_to_possible_relations(cls, selected_id: str) -> None:
        """
        Adds inactive relations to the list of possible relations for a given selected ID.
        This method retrieves inactive relations and checks if the selected ID matches either
        the source or target of those relations, adding them accordingly.

        :param selected_id: The identifier of the selected object for which inactive relations
                            are to be added.
        :type selected_id: str

        :return: None
        """

        inactive_relations = cls.get_inactive_aim_id_relations()
        for relation_instance in inactive_relations:
            source_id = relation_instance.bronAssetId.identificator
            target_id = relation_instance.doelAssetId.identificator
            if source_id == selected_id:
                cls.add_possible_relation_object(relation_object=relation_instance,
                                                 selected_object_id=source_id,
                                                 related_object_id=target_id)
            if target_id == selected_id:
                cls.add_possible_relation_object(relation_object=relation_instance,
                                                 selected_object_id=target_id,
                                                 related_object_id=source_id)

    @classmethod
    def are_possible_relations_to_other_class_types_collected_for(cls, typeURI: str):
        """
        Checks if possible relations for a given class type URI have been collected.
        This method verifies the presence of the type URI in the internal dictionary of
        possible relations.

        :param typeURI: The URI of the class type to check for collected relations.
        :type typeURI: str

        :return: True if possible relations have been collected for the specified type URI, otherwise False.
        :rtype: bool
        """

        return typeURI in cls.possible_relations_per_class_dict.keys()

    @classmethod
    def collect_possible_relations_to_class_types_from(cls, selected_object:RelationInteractor):
        """
        Collects possible relations for a specified selected object based on its type.
        This method populates the internal dictionary of possible relations, considering both
        internal and external objects, and handles exceptions during the collection process.

        :param selected_object: The object from which possible relations are to be collected.
        :type selected_object: RelationInteractor

        :return: None
        """
        selected_typeURI:str = selected_object.typeURI
        log_typeURI = RelationChangeHelpers.get_abbreviated_typeURI(selected_typeURI)
        OTLLogger.logger.debug(f"Execute RelationChangeDomain.collect_possible_relations_to_class_types_from({log_typeURI}) for project {global_vars.current_project.eigen_referentie}",
                               extra={"timing_ref": f"collect_possible_relations_classes"})

        if selected_typeURI.startswith("https://lgc"): # selected_object is legacy
            cls.possible_relations_per_class_dict[selected_typeURI] = (
                cls.get_hoortbij_relaties_from_legacy_asset(selected_typeURI))
        else:
            if (cls.external_objects or
                cls.agent_objects or
                cls.search_full_OTL_mode):
                # this is the long search, but it includes relations with external assets
                cls.possible_relations_per_class_dict[selected_typeURI] = (
                    RelationChangeDomain.get_all_concrete_relation_from_full_model(
                        selected_object=selected_object))
            else:
                try:
                    cls.possible_relations_per_class_dict[selected_typeURI] = \
                        cls.collector.find_all_concrete_relations(objectUri=selected_typeURI,
                                                              allow_duplicates=False)
                except ValueError as e:
                    OTLLogger.logger.debug(f"Didn't find relations in subset:\n {e}")
                    cls.possible_relations_per_class_dict[selected_typeURI] = (
                        RelationChangeDomain.get_all_concrete_relation_from_full_model(
                            selected_object=selected_object))

        relation_count = len(cls.possible_relations_per_class_dict[selected_object.typeURI])
        OTLLogger.logger.debug(
            f"Execute RelationChangeDomain.collect_possible_relations_to_class_types_from({log_typeURI}) for project {global_vars.current_project.eigen_referentie} ({relation_count} relations)",
            extra={"timing_ref": f"collect_possible_relations_classes"})

    @classmethod
    def get_hoortbij_relaties_from_legacy_asset(cls, selected_typeURI):
        all_concrete_classes = []
        all_classes_in_OTL = get_hardcoded_class_dict()
        for otl_class in all_classes_in_OTL.keys():
            if not otl_class.startswith("https://lgc"):
                concrete_classes = [non_legacy_class for non_legacy_class in
                                    get_concrete_subclasses_from_class_dict(otl_class) if
                                    not non_legacy_class.startswith("https://lgc")]
                all_concrete_classes.extend(concrete_classes)
        all_concrete_classes = list(set(all_concrete_classes))
        possible_relations = [
            RelationChangeDomain.create_hoortbij_OSLORelatie_with_legacy_class(selected_typeURI,
                                                                               concrete_class) for concrete_class in all_concrete_classes]
        return possible_relations

    @classmethod
    def create_hoortbij_OSLORelatie_with_legacy_class(cls, legacy_class_typeURI,
                                                      OTL_class_typeURI):
        return OSLORelatie(
            bron_overerving="",
            doel_overerving="",
            bron_uri=OTL_class_typeURI,
            doel_uri=legacy_class_typeURI,
            objectUri="https://wegenenverkeer.data.vlaanderen.be/ns/onderdeel#HoortBij",
            richting="Source -> Destination",
            deprecated_version="",
            usagenote="")

    @classmethod
    def get_all_concrete_relation_from_full_model(cls, selected_object:RelationInteractor):
        """
        Retrieves all concrete relations from the full model for a specified selected object.
        This method processes the relations and returns them in a structured format suitable for
        further use.

        :param selected_object: The object for which all concrete relations are to be retrieved.
        :type selected_object: RelationInteractor

        :return: A list of concrete OSLORelatie objects representing the relations.
        :rtype: list[OSLORelatie]
        """

        # noinspection PyProtectedMember
        all_relations = selected_object._get_all_concrete_relations()
        concrete_OSLO_relations: list[OSLORelatie] = [OSLORelatie(
            bron_overerving="",
            doel_overerving="",
            bron_uri=concrete_relation[0],
            doel_uri=concrete_relation[2],
            objectUri=concrete_relation[1],
            richting=concrete_relation[3],
            deprecated_version=concrete_relation[4],
            usagenote="") for concrete_relation in all_relations]
        return concrete_OSLO_relations

    @classmethod
    def get_same_relations_in_list(cls, relation_list: list[RelatieObject],
                                   relation_def: OSLORelatie,
                                   selected_object: RelationInteractor,
                                   related_object: RelationInteractor,
                                   reverse: bool = False):
        """
        Retrieves all relations from a given list that match a specified relation definition.
        This method checks for existing relations that are equivalent to the provided relation
        definition, considering the selected and related objects, and can account for reverse
        relationships.

        :param relation_list: A list of relation objects to search through.
        :type relation_list: list[RelatieObject]

        :param relation_def: The relation definition to match against.
        :type relation_def: OSLORelatie

        :param selected_object: The selected object involved in the relation.
        :type selected_object: RelationInteractor

        :param related_object: The related object involved in the relation.
        :type related_object: RelationInteractor

        :param reverse: A boolean indicating whether to check for reverse relations. Defaults to False.
        :type reverse: bool, optional

        :return: A list of relations that match the specified relation definition.
        :rtype: list
        """

        return [relation for relation in relation_list
                if cls.is_same_relation(existing_relation=relation, relation_def=relation_def,
                                        selected=selected_object, related=related_object,
                                        reverse=reverse)
                ]


    @classmethod
    def is_same_relation(cls, existing_relation: RelatieObject, relation_def: OSLORelatie,
                         selected:RelationInteractor, related: RelationInteractor,
                         reverse: bool = False) -> bool:
        """
        Determines whether an existing relation matches a specified relation definition
        based on the selected and related objects. This method checks the identifiers of the
        existing relation against the provided relation definition and can account for both standard
        and reverse relationships.

        :param existing_relation: The relation object to compare against the relation definition.
        :type existing_relation: RelatieObject

        :param relation_def: The relation definition to match.
        :type relation_def: OSLORelatie

        :param selected: The selected object involved in the relation.
        :type selected: RelationInteractor

        :param related: The related object involved in the relation.
        :type related: RelationInteractor

        :param reverse: A boolean indicating whether to check for reverse relations. Defaults to False.
        :type reverse: bool, optional

        :return: True if the existing relation matches the relation definition, otherwise False.
        :rtype: bool
        """

        existing_source_id:str = existing_relation.bronAssetId.identificator
        existing_target_id:str  = existing_relation.doelAssetId.identificator
        related_id:str = RelationChangeHelpers.get_corrected_identificator(related)
        selected_id:str = RelationChangeHelpers.get_corrected_identificator(selected)

        if existing_relation.typeURI != relation_def.objectUri:
            return False
        if relation_def.richting == "Unspecified":
            return ((existing_source_id == related_id and existing_target_id == selected_id) or
                    (existing_source_id == selected_id and existing_target_id == related_id))
        else:
            return ((existing_source_id == related_id  and existing_target_id == selected_id)
                    if reverse else
                    (existing_source_id == selected_id and existing_target_id == related_id))

    @classmethod
    def add_relation_between(cls, relation: OSLORelatie, selected_object: RelationInteractor,
                             related_object: RelationInteractor, reverse: bool = False) -> None:
        """
        Adds a relation between two specified objects, considering the direction of the relation.
        This method creates a relation object based on the provided relation definition and adds it
        to the internal dictionary of possible relations if it does not already exist.

        :param relation: The relation definition that describes how the objects are related.
        :type relation: OSLORelatie

        :param selected_object: The first object involved in the relation.
        :type selected_object: RelationInteractor

        :param related_object: The second object involved in the relation.
        :type related_object: RelationInteractor

        :param reverse: A boolean indicating whether to create the relation in the reverse direction. Defaults to False.
        :type reverse: bool, optional

        :return: None
        """

        selected_object_id:str = RelationChangeHelpers.get_corrected_identificator(selected_object)
        related_object_id:str = RelationChangeHelpers.get_corrected_identificator(related_object)

        # the relation object is reversed from the wrong direction to the right direction
        # noinspection PyUnusedLocal
        relation_object = None
        if reverse:
            relation_object = cls.create_relation_object(OSLO_relation=relation,
                                                         source_object=related_object,
                                                         target_object= selected_object)
        else:
            relation_object = cls.create_relation_object(OSLO_relation=relation,
                                                         source_object=selected_object,
                                                         target_object=related_object)

        if not relation_object:
            return

        if (selected_object_id in cls.possible_object_to_object_relations_dict.keys() and
                related_object_id in cls.possible_object_to_object_relations_dict[
                    selected_object_id].keys() and
                cls.get_same_relations_in_list(
                    cls.possible_object_to_object_relations_dict[selected_object_id][
                        related_object_id],
                    relation_def=relation,
                    selected_object=selected_object,
                    related_object=related_object,
                    reverse=reverse)):
            return

        cls.add_possible_relation_object(relation_object, selected_object_id, related_object_id)

    @classmethod
    def add_possible_relation_object(cls, relation_object, selected_object_id, related_object_id) \
            -> None:
        """
        Adds a relation object to the internal dictionary of possible relations between two
        specified objects. This method updates the dictionary to either append the relation to an
        existing list or create a new entry if no relations currently exist.

        :param relation_object: The relation object to be added.
        :type relation_object: Any

        :param selected_object_id: The identifier of the first object involved in the relation.
        :type selected_object_id: str

        :param related_object_id: The identifier of the second object involved in the relation.
        :type related_object_id: str

        :return: None
        """

        if selected_object_id in cls.possible_object_to_object_relations_dict:
            if related_object_id in cls.possible_object_to_object_relations_dict[
                selected_object_id]:
                cls.possible_object_to_object_relations_dict[selected_object_id][
                    related_object_id].append(relation_object)
            else:
                cls.possible_object_to_object_relations_dict[selected_object_id][
                    related_object_id] = [relation_object]
        else:
            cls.possible_object_to_object_relations_dict[selected_object_id] = {
                related_object_id: [relation_object]}


    @classmethod
    def filter_out(cls, object_to_filter_for: RelationInteractor)-> Callable:
        """
        Creates a filter function that excludes a specified object from a list of related objects.
        This method returns a callable that can be used to filter out the object to ensure it is
        not included in subsequent operations.

        :param object_to_filter_for: The object that should be excluded from the filtering process.
        :type object_to_filter_for: RelationInteractor

        :return: A callable function that takes a related object and returns True
                 if it should be included, or False if it should be excluded.
        :rtype: Callable
        """

        def filter_func(related_object: RelationInteractor):
            return (RelationChangeHelpers.get_corrected_identificator(object_to_filter_for) !=
                    RelationChangeHelpers.get_corrected_identificator(related_object))

        return filter_func


    @classmethod
    def get_screen(cls) -> RelationChangeScreenInterface:
        return global_vars.otl_wizard.main_window.step3_relations

    @classmethod
    def create_relation_object(cls,
                               OSLO_relation: OSLORelatie,
                               source_object: RelationInteractor,
                               target_object: RelationInteractor) -> Optional[RelatieObject]:
        """
        Creates a relation object based on the provided OSLO relation and the specified source and
        target objects. This method attempts to instantiate the relation object and assigns it
        relevant properties, logging any errors encountered during the process.

        :param OSLO_relation: The OSLO relation definition used to create the relation object.
        :type OSLO_relation: OSLORelatie

        :param source_object: The object that serves as the source in the relation.
        :type source_object: RelationInteractor

        :param target_object: The object that serves as the target in the relation.
        :type target_object: RelationInteractor

        :return: The created relation object if successful, or None if an error occurred.
        :rtype: Optional[RelatieObject]
        """

        try:
            relation_type = dynamic_create_type_from_uri(OSLO_relation.objectUri)
            # noinspection PyTypeChecker
            relation_object = RelationCreator.create_relation(relation_type= relation_type,
                                                              source=source_object,
                                                              target=target_object)
            relation_object.assetId.toegekendDoor = global_vars.external_toegekendDoor_label
        except Exception as e:
            source_id = None
            source_type = None
            if source_object:
                source_id = RelationChangeHelpers.get_corrected_identificator(
                    otl_object=source_object)
                source_type = RelationChangeHelpers.get_abbreviated_typeURI(
                        typeURI=source_object.typeURI, add_namespace=True)

            target_id = None
            target_type = None
            if target_object:

                target_id = RelationChangeHelpers.get_corrected_identificator(
                    otl_object=target_object)
                target_type = RelationChangeHelpers.get_abbreviated_typeURI(
                        target_object.typeURI, True)

            OTLLogger.logger.debug(
                f"Couldn't make relation between source_type:{source_type} source_id:{source_id} {OSLO_relation.richting} target_type:{target_type} target_id:{target_id} for relation typeURI {OSLO_relation.objectUri}: \n {e}")
            return None

        return relation_object

    @classmethod
    @async_save_assets
    async def add_multiple_possible_relations_to_existing_relations(cls, data_list):
        """
        Adds multiple possible relations to existing relations based on the provided data list.
        This method checks for specific attributes in the relations and either displays a dialog
        for user confirmation or directly updates the existing relations.

        :param data_list: A list of data objects containing source and target IDs along with their respective indices for the relations to be added.
        :type data_list: list

        :return: None
        """

        # sourcery skip: use-named-expression
        heeft_betrokkene_in_selection = [True for data in data_list
                                         if cls.possible_object_to_object_relations_dict[
                                             data.source_id][data.target_id][
                                             data.index].typeURI == 'https://wegenenverkeer.data.vlaanderen.be/ns/onderdeel#HeeftBetrokkene']

        if heeft_betrokkene_in_selection:
            data_list_and_relation_objects = [(data.source_id, data.target_id,
                                               cls.possible_object_to_object_relations_dict[
                                                   data.source_id][
                                                   data.target_id].pop(data.index)) for data in
                                              data_list]
            cls.get_screen().showMultiSelectionHeeftBetrokkeneAttributeDialogWindow(
                data_list_and_relation_objects=data_list_and_relation_objects)
        else:
            cls.last_added_to_existing = [
                RelationChangeDomain.add_possible_relation_to_existing_relations(
                    bron_asset_id=data.source_id,target_asset_id=data.target_id,
                    relation_object_index=data.index) for data in data_list]
        cls.update_frontend()

    @classmethod
    def add_possible_relation_to_existing_relations(cls, bron_asset_id:str, target_asset_id:str,
                                                    relation_object_index:int) -> RelatieObject:
        """
        Adds a possible relation to the existing relations using the specified source and target
        asset IDs along with the index of the relation object. This method retrieves the relation
        object from the internal dictionary, adds it to the existing relations, and returns the
        relation object.

        :param bron_asset_id: The identifier of the source asset for the relation.
        :type bron_asset_id: str

        :param target_asset_id: The identifier of the target asset for the relation.
        :type target_asset_id: str

        :param relation_object_index: The index of the relation object to be added.
        :type relation_object_index: int

        :return: The relation object that was added to the existing relations.
        :rtype: RelatieObject
        """

        relation_object = cls.possible_object_to_object_relations_dict[bron_asset_id][
            target_asset_id].pop(relation_object_index)

        cls.add_relation_object_to_existing_relations(relation_object=relation_object)

        return relation_object

    @classmethod
    def add_relation_object_to_existing_relations(cls, relation_object: RelatieObject):
        """
        Adds a relation object to the list of existing relations and marks it as active.
        This method also updates the user interface to reflect the addition by expanding the folder
        corresponding to the relation type.

        :param relation_object: The relation object to be added to the existing relations.
        :type relation_object: RelatieObject

        :return: None
        """
        global_vars.current_project.visualisation_uptodate.insert_relation(relation_object)
        relation_object.isActief = True
        cls.existing_relations.append(relation_object)
        cls.get_screen().expand_existing_relations_folder_of(
            relation_typeURI=relation_object.typeURI)



    @classmethod
    def update_frontend(cls):
        """
        Updates the user interface to reflect the current state of objects and relations.
        This method populates the object list and existing relations list, and sets the possible
        relations for the selected object.

        :return: None
        """
        OTLLogger.logger.debug("Execute RelationChangeDomain.update_frontend",
                               extra={"timing_ref": f"update_frontend"})
        cls.update_frontend_objects()
        cls.update_frontend_existing_relations()

        if cls.selected_object:
            create_task_reraise_exception(cls.set_possible_relations(selected_object=cls.selected_object))

        global_vars.otl_wizard.main_window.step3_visuals.update_only_legend()

        OTLLogger.logger.debug("Execute RelationChangeDomain.update_frontend",
                               extra={"timing_ref": f"update_frontend"})

    @classmethod
    def update_frontend_objects(cls):
        cls.get_screen().fill_object_list(objects=cls.shown_objects)

    @classmethod
    def update_frontend_existing_relations(cls):
        cls.get_screen().fill_existing_relations_list(relations_objects=cls.existing_relations,
                                                      last_added=cls.last_added_to_existing)

    @classmethod
    def update_frontend_possible_relations(cls):
        if cls.selected_object:
            selected_id = RelationChangeHelpers.get_corrected_identificator(otl_object=cls.selected_object)
            possible_relations_for_this_object = cls.get_possible_relations_for(
                selected_id=selected_id)

            cls.get_screen().fill_possible_relations_list(source_object=cls.selected_object,
                                                      relations=possible_relations_for_this_object,
                                                      last_added=cls.last_added_to_possible)

    @classmethod
    def set_selected_object_from_map(cls, identificator):
        cls.set_selected_object(cls.get_object(identificator))
        cls.get_screen().set_selected_object(identificator)

    @classmethod
    @async_save_assets
    async def remove_multiple_existing_relations(cls, indices: list[int]) -> None:
        """
        Removes multiple existing relations based on the provided indices. This method updates the
        internal state by removing the specified relations and refreshes the user interface to
        reflect the changes.

        :param indices: A list of indices corresponding to the existing relations to be removed.
        :type indices: list[int]

        :return: None
        """

        cls.last_added_to_possible = [cls.remove_existing_relation(index) for index in indices]

        cls.update_frontend()

    @classmethod
    def remove_existing_relation(cls, index: int) -> RelatieObject:
        """
        Removes an existing relation from the list based on the specified index. This method marks
        the removed relation as inactive, updates the user interface to reflect the change, and
        returns the removed relation object.

        :param index: The index of the existing relation to be removed.
        :type index: int

        :return: The relation object that was removed from the existing relations.
        :rtype: RelatieObject
        """

        removed_relation = cls.existing_relations.pop(index)
        global_vars.current_project.visualisation_uptodate.remove_relation(removed_relation)
        # if the removed relation already had an AIM ID it is set to false and kept if not it is
        # removed and made again in the possible relations
        removed_relation.isActief = False
        cls.get_screen().expand_possible_relations_folder_of(
            relation_typeURI=removed_relation.typeURI)

        return removed_relation

    @classmethod
    async def select_existing_relation_indices(cls, indices: list[int]) -> None:
        """
        Selects existing relations based on the provided indices and updates the user interface
        accordingly. This method retrieves the last selected relation and populates the attribute
        field with its details, or clears the field if no indices are provided.

        :param indices: A list of indices corresponding to the existing relations to be selected.
        :type indices: list[int]

        :return: None
        """

        if not indices:
            cls.get_screen().fill_existing_relation_attribute_field(
                existing_relation_attribute_dict={})
            return

        last_index = indices[-1]
        last_selected_relation = cls.existing_relations[last_index]
        cls.get_screen().fill_existing_relation_attribute_field(
            existing_relation_attribute_dict=await DotnotationDictConverter.to_dict_async(
                last_selected_relation))

    @classmethod
    async def select_possible_relation_data(cls, selected_relations_data: list) -> None:
        """
        Selects and displays data for possible relations based on the provided selected relations
        data. This method updates the user interface with the attributes of the last selected
        relation partner asset, or clears the attribute field if no relations are selected.

        :param selected_relations_data: A list of data objects representing the selected possible
                                        relations.
        :type selected_relations_data: list

        :return: None
        """

        if not selected_relations_data:
            cls.get_screen().fill_possible_relation_attribute_field({})
            return
        last_selected_keys = selected_relations_data[-1]
        last_selected_relation = \
        cls.possible_object_to_object_relations_dict[last_selected_keys.source_id][
            last_selected_keys.target_id][last_selected_keys.index]

        last_selected_relation_partner_asset: RelationInteractor = RelationChangeDomain.get_object(
            identificator=last_selected_relation.doelAssetId.identificator)
        if last_selected_relation_partner_asset == cls.selected_object:
            last_selected_relation_partner_asset = RelationChangeDomain.get_object(
                identificator=last_selected_relation.bronAssetId.identificator)

        # noinspection PyTypeChecker
        cls.get_screen().fill_possible_relation_attribute_field(
            possible_relation_attribute_dict=await DotnotationDictConverter.to_dict_async(
                                             otl_object=last_selected_relation_partner_asset))

    @classmethod
    def get_internal_objects(cls) -> list[RelationInteractor]:
        """
        Retrieves the list of internal objects managed by the class.
        This method provides access to the internal state of the class, allowing other components
        to utilize the internal objects as needed.

        :return: A list of internal RelationInteractor objects.
        :rtype: list[RelationInteractor]
        """

        return cls.internal_objects

    @classmethod
    def get_persistent_relations(cls) -> list[RelatieObject]:
        """
        Retrieves a combined list of persistent relations, this contains all active relations but
        only those inactive relations that are in the DAVIE database.
        This method provides access to all relations that are currently stored, allowing
        for comprehensive management of relational data.

        :return: A list of persistent relations, including existing and inactive (DAVIE) relations.
        :rtype: list[RelatieObject]
        """

        return cls.existing_relations + cls.get_inactive_aim_id_relations()

    @classmethod
    def get_export_instances(cls) -> list[Union[RelatieObject, RelationInteractor]]:
        """
         Retrieves a combined list of instances that need to be included in an exported file,
         including both internal objects and persistent relations.
         This method provides a comprehensive view of all objects that can be exported,
         facilitating data management and export operations.

        :return: A list of export instances, which may include RelatieObject and RelationInteractor types.
        :rtype: list[Union[RelatieObject, RelationInteractor]]
        """

        return cls.internal_objects + cls.get_persistent_relations()

    @classmethod
    def get_quicksave_instances(cls) -> list[Union[RelatieObject, RelationInteractor]]:
        """
        Retrieves a combined list of  instances to make a quicksave file,
        including both shown objects and persistent relations.
        This method provides a convenient way to access all objects that are currently visible
        and can be quickly saved, enhancing data management efficiency.

        :return: A list of quicksave instances, which may include RelatieObject and RelationInteractor types.
        :rtype: list[Union[RelatieObject, RelationInteractor]]
        """

        return cls.shown_objects + cls.get_persistent_relations()

    @classmethod
    def get_visualisation_instances(cls)-> list[Union[RelatieObject, RelationInteractor]]:
        # cls.visualisation_uptodate = True
        return [asset for asset in RelationChangeDomain.get_quicksave_instances() if
         asset.isActief != False]

    @classmethod
    def apply_active_aim_id_relations(cls) -> None:
        """
        Applies active AIM ID relations (already in DAVIE) to the list of existing relations.
        This method iterates through the AIM ID relations and appends only those that are
        marked as active to the existing relations, ensuring that the current state reflects only
        the relevant relations.

        :return: None
        """
        
        for relation_instance in cls.aim_id_relations:
            if relation_instance.isActief != False:
                cls.existing_relations.append(relation_instance)

    @classmethod
    def get_inactive_aim_id_relations(cls):
        """
        Retrieves a list of inactive AIM ID relations from the collection of AIM ID relations
        (already in DAVIE).
        This method filters the relations to return only those that are marked as inactive,
        providing a way to access non-active relations for further processing or analysis.

        :return: A list of inactive AIM ID relations.
        :rtype: list
        """

        return [relation_instance for relation_instance in cls.aim_id_relations if
                relation_instance.isActief == False]

    @classmethod
    def add_external_objects_to_shown_objects(cls):
        """
        Adds external objects to the list of shown objects.
        This method updates the internal state by extending the shown objects with external objects
        and marks that external objects have been added.

        :return: None
        """

        cls.shown_objects.extend(cls.external_objects)
        cls.regenerate_relation_types = True

    @classmethod
    def add_agent_objects_to_shown_objects(cls):
        """
        Adds agent objects to the list of shown objects. This method updates the internal state by
        extending the shown objects with agent objects and marks that external objects have been
        added.

        :return: None
        """
        cls.shown_objects.extend(cls.agent_objects)
        cls.regenerate_relation_types = True

    @classmethod
    @add_loading_screen
    @async_save_assets
    async def user_input_to_create_and_add_new_external_asset(cls, id_or_name: str, type_uri: str):
        cls.create_and_add_new_external_asset( id_or_name, type_uri)
        global_vars.current_project.visualisation_uptodate.set_clear_all(True)
        RelationChangeDomain.update_frontend()

    # noinspection PyUnresolvedReferences,PyTypeChecker
    @classmethod
    def create_and_add_new_external_asset(cls, id_or_name: str, type_uri: str):
        """
        Creates a new external asset based on the provided identifier or name and type URI,
        and adds it to the appropriate collections. This method distinguishes between agent objects
        and other asset types, updating the relevant lists and the user interface accordingly.

        :param id_or_name: The identifier or name to assign to the new external asset.
        :type id_or_name: str

        :param type_uri: The type URI that defines the kind of external asset to create.
        :type type_uri: str

        :return: None
        """

        new_external_object: RelationInteractor = dynamic_create_instance_from_uri(type_uri)
        if  hasattr(new_external_object,"agentId"):
            new_external_object.agentId.identificator = id_or_name
            cls.agent_objects.append(new_external_object)
        else:
            new_external_object.assetId.identificator = id_or_name
            new_external_object.assetId.toegekendDoor = global_vars.external_toegekendDoor_label
            cls.external_objects.append(new_external_object)

        cls.shown_objects.append(new_external_object)

        cls.regenerate_relation_types = True

        
    @classmethod
    def is_visualisation_uptodate(cls):
        return global_vars.current_project.visualisation_uptodate.is_uptodate()

    @classmethod
    def get_current_relation_change_screen_object_list_content_dict(cls):
        cls.map_uptodate = True
        return cls.get_screen().get_current_object_list_content_dict()

    @classmethod
    def get_map_screen(cls):
        return cls.get_screen().map_window
        # return global_vars.otl_wizard.main_window.step3_map

    @classmethod
    def get_shown_objects(cls):
        return cls.shown_objects

    @classmethod
    def set_search_full_OTL_mode(cls,state:bool) -> None:
        cls.search_full_OTL_mode = state
        cls.regenerate_relation_types = True




import asyncio
import sys
import traceback
from typing import Awaitable

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.GUI.screens.ErrorScreen import ErrorScreen


def excepthook(exc_type, exc_value, exc_tb):
    tb = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
    OTLLogger.logger.error("error caught!")
    OTLLogger.logger.error("error message: \n: " + tb)
    error_screen = ErrorScreen(tb)
    error_screen.show()
    # QApplication.quit()


def create_task_reraise_exception(awaitable: Awaitable) -> asyncio.Task:
    async def _reraise_exception(awaitable):
        try:
            return await awaitable
        except Exception as e:
            exc_type, exc_value, exc_tb = sys.exc_info()
            # Call your excepthook
            excepthook(exc_type, exc_value, exc_tb)

    event_loop = asyncio.get_event_loop()
    return  event_loop.create_task(_reraise_exception(awaitable))


from pathlib import Path

from otlmow_gui.Domain.enums import Language
from otlmow_gui.Domain.Settings import Settings


class GlobalTranslate:
    instance = None

    def __init__(self,settings:dict,lang_dir:Path):
        self._ = Settings.return_language(lang_dir,Language[settings["language"]])
        GlobalTranslate.instance = self

    @classmethod
    def _(cls,key:str) -> str:
        return cls.instance._(key)

    def get_all(self):
        return self._


from abc import abstractmethod

from PyQt6.QtWidgets import QWidget

from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.GUI.screens.Screen import Screen


class HomeScreenInterface(Screen):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def create_main_content_ui(self) -> None:
        pass
    @abstractmethod
    def fill_table(self,projects: [Project]) -> None:
        pass

    @abstractmethod
    def draw_search_bar(self) -> QWidget:
        pass

    @abstractmethod
    def create_input_field(self) -> None:
        pass

    @abstractmethod
    def remove_table_row(self, row_index: int) -> None:
        pass




from PyQt6.QtCore import Qt, QModelIndex
from PyQt6.QtGui import QStandardItemModel, QStandardItem
from PyQt6.QtWidgets import QTreeView, QVBoxLayout


class FolderTreeView(QTreeView):
    def __init__(self):
        super().__init__()

        # Create the QStandardItemModel
        self.model = QStandardItemModel()
        # self.model.setHorizontalHeaderLabels(["Folders and Items"])

        # Add custom folders and items
        # self.add_folder_with_items("Folder 1", ["Item 1.1", "Item 1.2", "Item 1.3"])
        # self.add_folder_with_items("Folder 2", ["Item 2.1", "Item 2.2"])
        # self.add_folder_with_items("Folder 3", ["Item 3.1", "Item 3.2", "Item 3.3", "Item 3.4"])

        self.setModel(self.model)


        self.setStyleSheet(
            """
             QHeaderView::section
            {
                    height: 14px;
                    font-size: 14px;
                    font-family: Roboto;
                    font-style: bold;
                    padding: 0px 0px 5px 5px;
                    margin: 0px;
            }
            """
        )
        # self.header().hide()
        # self.expandAll()

    def resize_listener(self, event):
        self.resize_columns()

    def resize_columns(self,multi_col_list=True):
        half_width = int(self.width()/ 2)
        # if len(self.labels) > 1:
        if multi_col_list and self.columnWidth(0) > half_width:
            self.setColumnWidth(0, half_width)
            # self.setColumnWidth(1, half_width)

    def add_folder_with_items(self, folder_name, items):
        # Create a folder item (top-level item)
        folder_item = QStandardItem(folder_name)
        folder_item.setEditable(False)  # Make the folder name non-editable
        folder_item.setSelectable(False)  # Optional: make the folder itself non-selectable
        folder_item.setToolTip("")
        folder_item.setData(None, Qt.ItemDataRole.ToolTipRole)

        # Add items to the folder
        for item_name in items:
            item = QStandardItem(item_name)
            item.setEditable(False)  # Make the item name non-editable
            item.setToolTip("")
            item.setData(None, Qt.ItemDataRole.ToolTipRole)
            folder_item.appendRow(item)

        # Add the folder to the model
        self.model.appendRow(folder_item)

    def clear(self):
        self.model.clear()

    def addItem(self,folder_item):
        self.model.appendRow(folder_item)

    def event(self, event):
        if event.type() == event.Type.ToolTip:
            return False  # Ignore tooltip events
        return super().event(event)

    def toggle_expand_state_of_item_at_row(self, table_coord: QModelIndex):
        if self.isExpanded(table_coord):
            self.collapse(table_coord)
            return False

        self.expand(table_coord)
        return True


from pathlib import Path
from typing import List, Iterable, Optional, cast, Union

from openpyxl.reader.excel import load_workbook
from otlmow_converter.Exceptions.ExceptionsGroup import ExceptionsGroup

from otlmow_model.OtlmowModel.BaseClasses.OTLObject import OTLObject, \
    dynamic_create_instance_from_uri
from otlmow_model.OtlmowModel.BaseClasses.RelationInteractor import RelationInteractor
from otlmow_model.OtlmowModel.Classes.Agent import Agent
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.AIMObject import AIMObject
from otlmow_model.OtlmowModel.Classes.ImplementatieElement.RelatieObject import RelatieObject
from otlmow_model.OtlmowModel.Helpers import OTLObjectHelper, RelationValidator

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.project.ProjectFile import ProjectFile
from otlmow_gui.Domain.util.CombineAssetHelper import CombineAssetHelper
from otlmow_gui.Domain.util.Helpers import Helpers
from otlmow_gui.Domain.util.SDFHandler import SDFHandler
from otlmow_gui.Domain.logger.OTLLogger import OTLLogger
from otlmow_gui.Domain.project.Project import Project
from otlmow_gui.Domain.step_domain.RelationChangeDomain import RelationChangeDomain, \
    async_save_assets
from otlmow_gui.Domain.enums import FileState
from otlmow_gui.Exceptions.NoIdentificatorError import NoIdentificatorError
from otlmow_gui.Exceptions.RelationHasInvalidTypeUriForSourceAndTarget import \
    RelationHasInvalidTypeUriForSourceAndTarget
from otlmow_gui.Exceptions.RelationHasNonExistingTypeUriForSourceOrTarget import \
    RelationHasNonExistingTypeUriForSourceOrTarget
from otlmow_gui.GUI.dialog_windows.LoadingImageWindow import add_loading_screen_no_delay
from otlmow_gui.GUI.screens.RelationChange_elements.RelationChangeHelpers import RelationChangeHelpers
from otlmow_gui.GUI.translation.GlobalTranslate import GlobalTranslate




class InsertDataDomain:
    """
    Handles operations related to inserting and validating documents wih objects.

    This class provides methods for managing the insertion of data, including
    loading and validating documents, managing relations, and synchronizing
    data between the backend and frontend. It also includes functionality for
    handling exceptions related to data integrity.

    :param None:
    :attributes:
        project (optional): The current project associated with the data insertion.
        collector (OSLOCollector): Collector for managing OSLO data.
        selected_object (optional): The currently selected object in the UI.
        shown_objects (list): List of objects currently displayed in the UI.
        internal_objects (list): List of internal objects related to the project.
        external_objects (list): List of external objects related to the project.
        existing_relations (list): List of existing relations between objects.
        possible_relations_per_class_dict (dict): Dictionary of possible relations per class.
        possible_object_to_object_relations_dict (dict): Dictionary of possible relations between objects.
        aim_id_relations (list): List of relations identified by AIM IDs.
        no_id_count (int): Counter for objects without identifiers.
        external_object_added (bool): Flag indicating if an external object has been added.
    """

    @classmethod
    def init_static(cls,force_refresh = False):
        """
        Initializes static resources for the InsertDataDomain class.

        This method synchronizes the backend documents with the frontend, ensuring that
        the data displayed to the user is up-to-date and consistent with the backend state.

        :param cls: The class itself.
        :returns: None
        """
        if not Helpers.all_OTL_asset_types_dict or force_refresh:
            Helpers.create_external_typeURI_options()

        cls.get_screen().clear_feedback()
        cls.update_frontend()

    @classmethod
    async def check_document(cls, doc_location: Union[str, Path], delimiter: str=";") -> Iterable[OTLObject]:
        """
        Checks a document and converts it into a list of OTL objects.

        This method takes the location of a document and uses the OtlmowConverter
        to convert the contents of the file into a collection of OTL objects,
        including tab information for each object.

        :param cls: The class itself.
        :param doc_location: The path to the document to be checked.
        :type doc_location: str
        :returns: An iterable of OTL objects extracted from the document.
        """
        if isinstance(doc_location,str):
            doc_location_path = Path(doc_location)
        else:
            doc_location_path = doc_location

        exception_group = None
        try:
            if doc_location_path.suffix in ['.xls', '.xlsx']:
                temp_path = InsertDataDomain.remove_dropdown_values_from_excel(doc=doc_location_path)
                assets, exception_group =  await Helpers.converter_from_file_to_object_async(
                    file_path=temp_path,include_tab_info=True)

            elif doc_location_path.suffix == '.sdf':
                # SDF files will make multiple CSV files, one for each class
                temp_path_list = InsertDataDomain.create_temporary_SDF_conversion_to_CSV_files(
                    sdf_filepath=doc_location_path)

                assets = []
                sdf_exception_list = []
                for temp_path in temp_path_list:
                    assets_subset, exception_group_subset = await Helpers.converter_from_file_to_object_async(
                        file_path=Path(temp_path),
                        delimiter=",",
                        include_tab_info=True,
                        allow_non_otl_conform_attributes=True)

                    for asset in assets_subset:
                        if hasattr(asset,"FeatId"):
                            delattr(asset,"FeatId")

                    assets.extend(assets_subset)
                    if exception_group_subset is not None:
                        sdf_exception_list.extend(exception_group_subset.exceptions)

                exception_group = ExceptionsGroup(
                    message=f'Failed to create objects from CSV file (SDF are split into CSV\'s) {doc_location_path}')
                for exception in sdf_exception_list:
                    exception_group.add_exception(exception)
            else:
                assets, exception_group = await Helpers.converter_from_file_to_object_async(
                    file_path=doc_location_path,include_tab_info=True )

            # second checks done by the GUI
            if exception_group is None:
                exception_group = ExceptionsGroup(
                    message=f'Failed to create objects from {doc_location_path.suffix.upper()} file {doc_location_path}')

            cls.check_for_empty_identificators(assets=assets, exception_group=exception_group)

        except ExceptionsGroup as group:
            exception_group = group
            assets = group.objects

        return assets,exception_group

    @classmethod
    def remove_dropdown_values_from_excel(cls, doc: Path) -> Path:
        """
        Removes specific dropdown sheets from an Excel document.

        This method loads an Excel workbook from the specified document path,
        removes the sheets named 'Keuzelijsten' and 'dropdownvalues' if they exist,
        and saves the modified workbook to a temporary path. It returns the path
        of the modified workbook.

        :param cls: The class itself.
        :param doc: The path to the Excel document to be modified.
        :type doc: Path
        :returns: The path to the modified Excel document.
        """

        OTLLogger.logger.debug("starting excel changes")
        wb = load_workbook(doc)
        temp_path = Helpers.create_temp_path(path_to_template_file_and_extension=doc)
        if 'Keuzelijsten' in wb.sheetnames:
            wb.remove(wb['Keuzelijsten'])
        if 'dropdownvalues' in wb.sheetnames:
            wb.remove(wb['dropdownvalues'])

        wb.save(temp_path)
        return temp_path

    @classmethod
    def create_temporary_SDF_conversion_to_CSV_files(cls, sdf_filepath: Path) -> list[str]:

        temp_csv_path = Helpers.create_temp_path(path_to_template_file_and_extension=sdf_filepath).with_suffix(".csv")
        temp_csv_path_str_list = SDFHandler.convert_SDF_to_CSV(sdf_filepath=sdf_filepath,csv_output_path=temp_csv_path)

        return temp_csv_path_str_list


    @classmethod
    def add_template_file_to_project(cls, filepath: Path, project: Project,
                                     state: FileState) -> None:
        """Adds a template file to the specified project.

        This method checks the file extension of the provided template file and, if it is an Excel file,
        removes any dropdown values before adding it to the project. It then copies the file to the project
        and synchronizes the backend documents with the frontend.

        :param cls: The class itself.
        :param filepath: The path to the template file to be added.
        :type filepath: Path
        :param project: The project to which the template file will be added.
        :type project: Project
        :param state: The state of the file being added, indicating its status.
        :type state: FileState
        :returns: None
        """

        if Path(filepath).suffix in ['.xls', '.xlsx']:
            filepath = cls.remove_dropdown_values_from_excel(doc=filepath)

        project.copy_and_add_project_file(file_path=filepath, state=state)
        cls.update_frontend()

    @classmethod
    def return_temporary_path(cls, file_path: Path) -> Path:
        """
        Returns a temporary path based on the file type of the provided file path.

        This method checks the file extension of the given file path and returns a temporary path
        accordingly. If the file is an Excel file, it removes dropdown values; if it is a CSV file,
        it creates a temporary path for the file.

        :param cls: The class itself.
        :param file_path: The path to the file for which a temporary path is requested.
        :type file_path: Path
        :returns: The temporary path for the processed file.
        :rtype: Path
        """

        if Path(file_path).suffix in ['.xls', '.xlsx']:
            return cls.remove_dropdown_values_from_excel(doc=file_path)
        elif Path(file_path).suffix == '.csv':
            return Helpers.create_temp_path(path_to_template_file_and_extension=file_path)

    @classmethod
    def add_files_to_backend_list(cls, files: list[str],
                                  states: Optional[list[FileState]] = None) -> None:
        """
        Adds a list of files to the backend project with specified states.

        This method takes a list of file paths and an optional list of states,
        adding each file to the current project in the backend. If no states are
        provided, it defaults to a warning state for each file, and updates the
        file list in the user interface after the operation.

        :param cls: The class itself.
        :param files: A list of file paths to be added to the backend project.
        :type files: list[str]
        :param states: An optional list of states corresponding to each file.
        :type states: Optional[list[FileState]]
        :returns: None
        """

        if states is None:
            states = [FileState.WARNING for _ in range(len(files))]

        for i in range(len(files)):
            InsertDataDomain.add_template_file_to_project(project=global_vars.current_project,
                                                          filepath=Path(files[i]),
                                                          state=states[i])

        cls.get_screen().update_file_list()

    @classmethod
    def update_frontend(cls) -> bool:
        """
        Synchronizes the backend project files with the frontend display.

        This method clears the current list of project files in the frontend and
        populates it with the saved project files from the backend. It checks the
        state of each file and returns a boolean indicating whether all files are valid.

        :param cls: The class itself.
        :returns: True if all files are valid; otherwise, False.
        :rtype: bool
        """
        cls.get_screen().project_files_overview_field.clear()
        cls.get_screen().add_file_overview_placeholder_to_front_end_list()
        files_in_project = global_vars.current_project.get_saved_projectfiles()
        all_valid = False
        if files_in_project:
            cls.get_screen().project_files_overview_field.clear()

            all_valid = True
            for item in global_vars.current_project.get_saved_projectfiles():
                cls.get_screen().add_file_to_frontend_list(file=item.file_path,asset_state=item.state)
                if item.state != FileState.OK:
                    all_valid = False

        cls.get_screen().update_control_button_state()
        global_vars.otl_wizard.main_window.enable_steps()
        return all_valid

    @classmethod
    def delete_backend_document(cls, item_file_path: str) -> None:
        """Deletes a document from the backend project.

        This method removes the specified file from the current project's list of
        project files and synchronizes the backend documents with the frontend
        display to reflect the changes.

        :param cls: The class itself.
        :param item_file_path: The path of the file to be deleted from the backend.
        :type item_file_path: str
        :returns: None

        """
        global_vars.current_project.remove_project_file(Path(item_file_path))
        cls.get_screen().clear_feedback()

        InsertDataDomain.update_frontend()

    @classmethod
    def check_current_project_project_files_existence(cls) -> list[ProjectFile]:
        missing_project_files = global_vars.current_project.check_if_project_files_exist()

        if len(missing_project_files):
            global_vars.current_project.save_project_filepaths_to_file()
            cls.update_frontend()

        return missing_project_files
    @classmethod
    @add_loading_screen_no_delay
    @async_save_assets
    async def load_and_validate_documents(cls,**kwargs) -> tuple[list[dict], list]:
        """
        Loads and validates documents from the current project's saved files.

        This method iterates through the saved project files, attempting to load
        and validate each document. It collects any errors encountered during the
        process and returns a list of errors along with a list of successfully
        loaded objects.

        :param cls: The class itself.
        :returns: A tuple containing:
            - A list of error dictionaries, each containing an exception and the
              associated file path.
            - A list of successfully loaded objects.
        :rtype: tuple[list[dict],list]
        """
        OTLLogger.logger.debug(f"Executing InsertDataDomain.load_and_validate_documents() for project {global_vars.current_project.eigen_referentie}", extra={"timing_ref": f"validate_{global_vars.current_project.eigen_referentie}"})
        error_set: list[dict] = []
        relations_per_filepath_str_dict: dict[str,list[RelatieObject]] = {}
        assets_per_filepath_str_dict: dict[str, list[AIMObject]] = {}

        for project_file in global_vars.current_project.get_saved_projectfiles():
            file_path = project_file.file_path
            try:

                objects, exception_group = await cls.check_document(file_path)

                assets, relations = cls.divide_otl_objects(objects)

                if len(relations):
                    relations_per_filepath_str_dict[str(file_path)] = relations
                if len(assets):
                    # noinspection PyTypeChecker
                    assets_per_filepath_str_dict[str(file_path)] = assets

                if len(exception_group.exceptions) > 0:
                    raise exception_group

                project_file.state = FileState.OK

            except Exception as ex:
                error_set.append({"exception": ex, "path_str": file_path})
                project_file.state = FileState.ERROR

        total_assets = cls.flatten_list(objects_lists=list(assets_per_filepath_str_dict.values()))

        # check if there are relations without a bron/doel.TypeURI that cannot be tied to an asset
        for project_file in global_vars.current_project.get_saved_projectfiles():
            file_path = project_file.file_path
            file_path_str = str(file_path)

            if file_path_str not in relations_per_filepath_str_dict.keys():
                continue

            try:
                exception_group = ExceptionsGroup(
                    message=f'Failed to create objects from Excel file {file_path_str}')

                cls.check_for_invalid_relations(relations=relations_per_filepath_str_dict[file_path_str],
                                                ref_assets=total_assets, exception_group=exception_group)

                if len(exception_group.exceptions) > 0:
                    relations_per_filepath_str_dict.pop(file_path_str)
                    raise exception_group

                if project_file.state != FileState.ERROR:
                    project_file.state = FileState.OK
            except Exception as ex:
                error_set.append({"exception": ex, "path_str": file_path})
                project_file.state = FileState.ERROR

        if len(error_set):
            OTLLogger.logger.debug(
                f"Executing InsertDataDomain.load_and_validate_documents() for project {global_vars.current_project.eigen_referentie} (INVALID)",
                extra={"timing_ref": f"validate_{global_vars.current_project.eigen_referentie}"})
            # state can be changed to either OK or ERROR
            global_vars.current_project.save_project_filepaths_to_file()
            cls.update_frontend()
            return error_set, []

        objects_in_memory = cls.flatten_list(objects_lists=list(assets_per_filepath_str_dict.values()))
        # noinspection PyTypeChecker
        objects_in_memory.extend(cls.flatten_list(objects_lists=list(relations_per_filepath_str_dict.values())))

        try:
            objects_in_memory = await cls.combine_assets_wrapper(objects_in_memory)
        except ExceptionsGroup as ex:
            error_set.append({"exception": ex, "path_str": GlobalTranslate._("[FILE INFO NOT AVAILABLE]")})
            # the be safe set all document to invalid
            for project_file in global_vars.current_project.get_saved_projectfiles():
                project_file.state = FileState.ERROR

        # state can be changed to either OK or ERROR
        global_vars.current_project.save_project_filepaths_to_file()
        cls.update_frontend()

        # passing objects to other screens
        # global_vars.otl_wizard.main_window.step3_visuals.create_html(
        #     objects_in_memory=objects_in_memory)
        await RelationChangeDomain.set_instances(objects_list=objects_in_memory)
        # global_vars.otl_wizard.main_window.step3_visuals.reload_html()
        global_vars.current_project.visualisation_uptodate.set_clear_all(True)
        global_vars.current_project.save_visualisation_python_support_data()

        object_count = len(objects_in_memory)
        OTLLogger.logger.debug(
            f"Executing InsertDataDomain.load_and_validate_documents() for project {global_vars.current_project.eigen_referentie} ({object_count} objects)",
            extra={"timing_ref": f"validate_{global_vars.current_project.eigen_referentie}"})

        return error_set, objects_in_memory

    @classmethod
    async def combine_assets_wrapper(cls, objects_in_memory):

        objects_in_memory = CombineAssetHelper.combine_assets(objects_in_memory)

        return objects_in_memory

    @classmethod
    def divide_otl_objects(cls, objects):
        relations: list[RelatieObject] = []
        assets: list[OTLObject] = []
        for otl_object in objects:
            if OTLObjectHelper.is_relation(otl_object=otl_object):
                relations.append(otl_object)
            else:
                assets.append(otl_object)
        return assets, relations

    @classmethod
    def check_for_invalid_relations(cls, ref_assets: list[AIMObject], relations: list[RelatieObject],
                                    exception_group: ExceptionsGroup) -> None:
        """
        Checks for invalid relations among the provided assets.

        This method iterates through a list of assets and verifies the validity of
        relations defined within them. It adds any detected issues to the provided
        exception group, ensuring that all invalid relations are reported for further
        handling.

        :param cls: The class itself.
        :param ref_assets: A list of OTL objects to be checked for invalid relations.
        :type ref_assets: list[OTLObject]
        :param exception_group: A group to collect exceptions related to invalid relations.
        :type exception_group: ExceptionsGroup
        :returns: None
        """
        id_to_typeURI_dict: Optional[dict[str,str]] = None
        for relation in relations:

            bron_type_uri = Helpers.extract_corrected_relation_partner_typeURI(
                relation.bron.typeURI,
                relation.bronAssetId.identificator,
                id_to_typeURI_dict,
                ref_assets)

            if not bron_type_uri:
                ex = RelationHasNonExistingTypeUriForSourceOrTarget(
                    relation_type_uri=relation.typeURI,
                    relation_identificator=relation.assetId.identificator,
                    wrong_field="bron.typeURI",
                    wrong_value=None,
                    tab=RelationChangeHelpers.get_abbreviated_typeURI(
                        relation.typeURI, False))
                exception_group.add_exception(error=ex)

            doel_type_uri = Helpers.extract_corrected_relation_partner_typeURI(
                relation.doel.typeURI,
                relation.doelAssetId.identificator,
                id_to_typeURI_dict,
                ref_assets)

            if not doel_type_uri:
                ex = RelationHasNonExistingTypeUriForSourceOrTarget(
                    relation_type_uri=relation.typeURI,
                    relation_identificator=relation.assetId.identificator,
                    wrong_field="bron.typeURI",
                    wrong_value=None,
                    tab=RelationChangeHelpers.get_abbreviated_typeURI(
                        relation.typeURI, False))
                exception_group.add_exception(error=ex)

            if not bron_type_uri or not doel_type_uri:
                # the rest of the check will fail anyway
                continue

            has_valid_bron_typeURI:bool = InsertDataDomain.does_typeURI_exist(bron_type_uri)
            if not has_valid_bron_typeURI:
                ex = RelationHasNonExistingTypeUriForSourceOrTarget(
                    relation_type_uri=relation.typeURI,
                    relation_identificator=relation.assetId.identificator,
                    wrong_field="bron.typeURI",
                    wrong_value=bron_type_uri,
                    tab=RelationChangeHelpers.get_abbreviated_typeURI(
                        relation.typeURI,False))
                exception_group.add_exception(error=ex)

            has_valid_doel_typeURI: bool = InsertDataDomain.does_typeURI_exist(doel_type_uri)
            if not has_valid_doel_typeURI:
                ex = RelationHasNonExistingTypeUriForSourceOrTarget(
                    relation_type_uri=relation.typeURI,
                    relation_identificator=relation.assetId.identificator,
                    wrong_field="doel.typeURI",
                    wrong_value=doel_type_uri,
                    tab=RelationChangeHelpers.get_abbreviated_typeURI(
                        relation.typeURI,False))
                exception_group.add_exception(error=ex)


            if not has_valid_bron_typeURI or not has_valid_doel_typeURI:
                # the rest of the check will fail anyway
                continue

            try:
                is_valid_relation = RelationValidator.is_valid_relation(
                    relation_type=type(relation),
                    source_typeURI=bron_type_uri,
                    target_typeURI=doel_type_uri)
            except Exception as e:
                OTLLogger.logger.warning(e)
                is_valid_relation = False

            if not is_valid_relation:
                ex = cls.raise_wrong_doel_or_target(relation=relation,
                                                    tab=RelationChangeHelpers.get_abbreviated_typeURI(
                                                        relation.typeURI, False),
                                                    bron_type_uri=bron_type_uri,
                                                    doel_type_uri=doel_type_uri)
                exception_group.add_exception(error=ex)

    @classmethod
    def does_typeURI_exist(cls, bron_type_uri):
        return bron_type_uri in Helpers.all_OTL_asset_types_dict.values()



    @classmethod
    def raise_wrong_doel_or_target(cls, relation: RelatieObject, tab: str, bron_type_uri,
                                   doel_type_uri) -> RelationHasInvalidTypeUriForSourceAndTarget:
        """
        Creates an exception for an invalid relation with incorrect source or target type URIs.

        This method constructs and returns an instance of 
        RelationHasInvalidTypeUriForSourceAndTarget, populated with details about 
        the invalid relation, including the type URIs and the associated tab.

        :param doel_type_uri:
        :param bron_type_uri:
        :param cls: The class itself.
        :param relation: The relation object that contains the type URIs to be validated.
        :type relation: RelatieObject
        :param tab: The name of the tab where the relation is defined.
        :type tab: str
        :returns: An instance of RelationHasInvalidTypeUriForSourceAndTarget.
        :rtype: RelationHasInvalidTypeUriForSourceAndTarget
        """

        return RelationHasInvalidTypeUriForSourceAndTarget(
            relation_type_uri=relation.typeURI,
            relation_identificator=relation.assetId.identificator,
            wrong_field="bron.typeURI",
            wrong_value=bron_type_uri,
            wrong_field2="doel.typeURI",
            wrong_value2=doel_type_uri,
            tab=tab)

    @classmethod
    def flatten_list(cls, objects_lists: list[list[AIMObject]]) -> list[AIMObject]:
        """Flattens a list of lists into a single list of AIM objects.

        This method combines multiple lists of AIM objects into a single list, 
        making it easier to process all objects at once. It iterates through each 
        sublist and extends the main list with the objects contained in them.

        :param cls: The class itself.
        :param objects_lists: A list of lists containing AIM objects to be flattened.
        :type objects_lists: list[list[AIMObject]]
        :returns: A single list containing all AIM objects from the input lists.
        :rtype: list[AIMObject]
        """

        objects_in_memory: List[AIMObject] = []
        for objects_list in objects_lists:
            objects_in_memory.extend(objects_list)
        return objects_in_memory

    @classmethod
    def get_screen(cls):
        return global_vars.otl_wizard.main_window.step2

    @classmethod
    def detect_more_complex_target_or_source_typeURI_errors(cls, relation: RelatieObject) -> None:
        """
        @Unused
        Detects complex errors related to source or target type URIs in a relation.

        This method validates the relationship between the source and target assets, 
        checking for the existence of valid relations. If the relations are invalid, 
        it raises an appropriate error indicating the specific issue with the source 
        or target type URIs.

        :param cls: The class itself.
        :param relation: The relation object to be validated.
        :returns: None
        """

        if RelationValidator.is_valid_relation(relation_type=type(relation),
                                                   source_typeURI=relation.bron.typeURI,
                                                   target_typeURI=relation.doel.typeURI):
            return

        # RelationValidator.is_valid_relation doesn't say if bron or doel is wrong
        source_instance:RelationInteractor = dynamic_create_instance_from_uri(relation.bron.typeURI)
        concrete_source_relations = list(source_instance._get_all_concrete_relations())
        if concrete_source_relations_of_type_relation := {
            rel for rel in concrete_source_relations if rel[1] == relation.typeURI
        }:
            if  [
                rel
                for rel in concrete_source_relations_of_type_relation
                if rel[2] == relation.doel.typeURI]:
                OTLLogger.logger.debug("Error in logic")
            else:
                # source asset doesn't have this relation to target
                cls.raise_wrong_doel_or_target(relation=relation,
                                               tab=RelationChangeHelpers.get_abbreviated_typeURI(
                                                   typeURI=relation.typeURI,
                                                   add_namespace=False),
                                               bron_type_uri=relation.bron.typeURI,
                                               doel_type_uri=relation.doel.typeURI)
        else:
            # target asset has relation but not to source
            cls.raise_wrong_doel_or_target(relation=relation,
                                           tab=RelationChangeHelpers.get_abbreviated_typeURI(
                                               typeURI=relation.typeURI,
                                               add_namespace=False),
                                           bron_type_uri=relation.bron.typeURI,
                                           doel_type_uri=relation.doel.typeURI)

    @classmethod
    def check_for_empty_identificators(cls, assets: Iterable[OTLObject],
                                       exception_group: ExceptionsGroup) -> None:
        """Checks for empty identificators in the provided assets.

        This method iterates through a collection of assets and verifies that each asset 
        has a valid identificator. If an asset is found to be missing an identificator, 
        it adds an exception to the provided exception group.

        :param cls: The class itself.
        :param assets: An iterable collection of OTL objects to be checked for identificators.
        :type assets: Iterable[OTLObject]
        :param exception_group: A group to collect exceptions related to missing identificators.
        :type exception_group: ExceptionsGroup
        :returns: None
        """

        for asset in assets:
            identificator = None
            if asset.typeURI == 'http://purl.org/dc/terms/Agent':
                new_external_agent: Agent = cast(Agent, asset)
                identificator = new_external_agent.agentId.identificator

            else:
                new_external_asset: AIMObject = cast(AIMObject, asset)
                identificator = new_external_asset.assetId.identificator

            if not identificator:
                exception_group.add_exception(error=NoIdentificatorError(GlobalTranslate._ ,original_exception=ValueError(),tab=RelationChangeHelpers.get_abbreviated_typeURI(asset.typeURI,False)))



from typing import Optional

from otlmow_gui.GUI.dialog_windows.file_picker_dialog.LoadFilePickerDialog import LoadFilePickerDialog


class SubsetLoadFilePickerDialog(LoadFilePickerDialog):

    instance = None

    def __init__(self, language_settings,title:Optional[str]=None,name_filter:str=None, initial_file_path: Optional[str]=None):
        super().__init__(language_settings,title,name_filter, initial_file_path)

        if not title:
            self.setWindowTitle(self._("Selecteer subset"))
        if not name_filter:
            self.setNameFilter(self._("Database files (*.db)"))

        SubsetLoadFilePickerDialog.instance = self
        #so you can easily access this from UpsertProjectWindow and ChangeSubsetWindow


from pathlib import Path
import sys


class ProgramFileStructure:
    @classmethod
    def get_home_path(cls) -> Path:
        return Path.home()


    @classmethod
    def get_otl_wizard_work_dir(cls) -> Path:
        work_dir_path = cls.get_home_path() / 'OTLWizardProjects'
        if not work_dir_path.exists():
            work_dir_path.mkdir()
        return work_dir_path


    @classmethod
    def get_otl_wizard_projects_dir(cls) -> Path:
        projects_dir_path = cls.get_otl_wizard_work_dir() / 'Projects'
        if not projects_dir_path.exists():
            projects_dir_path.mkdir()
        return projects_dir_path

    @classmethod
    def get_otl_wizard_model_dir(cls) -> Path:
        model_dir_path = cls.get_otl_wizard_work_dir() / 'Model'
        if not model_dir_path.exists():
            model_dir_path.mkdir()

        return model_dir_path

    _dynamic_library_cache = {}

    @classmethod
    def get_dynamic_library_path(cls, library_name: str) -> Path:
        """Returns the resolved path to a dynamic library, using a cache for efficiency.

        This method attempts to resolve the path to the given library name using several strategies.
        The result is cached per library name for faster repeated lookups.

        Args:
            library_name (str): The name of the dynamic library.

        Returns:
            Path: The resolved path to the dynamic library.
        """
        if library_name in cls._dynamic_library_cache:
            return cls._dynamic_library_cache[library_name]

        dynamic_library_path = Path(library_name)
        if hasattr(sys, '_MEIPASS'):  # when in .exe file
            dynamic_library_path = Path(sys._MEIPASS, library_name)
        if not dynamic_library_path.exists():
            dynamic_library_path = Path('data', library_name)
        if not dynamic_library_path.exists():
            dynamic_library_path = Path('.').parent / library_name
        if not dynamic_library_path.exists():
            dynamic_library_path = Path(__file__).parent.parent / library_name
        if not dynamic_library_path.exists() and library_name == 'pyproject.toml':
            dynamic_library_path = Path(__file__).parent.parent.parent / library_name

        cls._dynamic_library_cache[library_name] = dynamic_library_path
        return dynamic_library_path


import logging
import os.path
from pathlib import Path

from PyQt6.QtWidgets import QFileDialog


class ExportToTemplateWindow(QFileDialog):

    previous_selected_directory = None

    def __init__(self):
        super().__init__()

        start_dir = str(Path.home() / "Documents")
        if ExportToTemplateWindow.previous_selected_directory:
            start_dir = str(ExportToTemplateWindow.previous_selected_directory)

        self.setModal(True)
        self.setDirectory(start_dir)

    def get_file_location(self) -> str:
        """
        Prompts the user to select a file location for saving.

        This function opens a file dialog that allows the user to choose a location to save a file,
        with options for Excel and CSV formats. It returns the selected file path or None if no
        file was selected.

        Args:
            self: The instance of the class.

        Returns:
            str: The path of the selected file, or None if no file was selected.
        """

        document_loc = self.getSaveFileName(
            filter="Excel files (*.xlsx);;CSV files (*.csv)")
        if document_loc == ('', ''):
            logging.debug("No file selected")
            return None
        elif document_loc:
            ExportToTemplateWindow.previous_selected_directory = os.path.dirname(str(document_loc[0]))

            return document_loc[0]





from PyQt6.QtWidgets import QDialog, QVBoxLayout, QLabel, QDialogButtonBox

from otlmow_gui.Domain import global_vars
from otlmow_gui.Domain.step_domain.InsertDataDomain import InsertDataDomain
from otlmow_gui.exception_handler.ExceptionHandlers import create_task_reraise_exception


class OverwriteGraphWarningWindow:

    def __init__(self, screen, language_settings):
        self.screen =screen
        self._ = language_settings
        self.init_ui()

    def init_ui(self):
        dialog = QDialog()
        dialog.setModal(True)
        dialog.setWindowTitle(self._("overwrite_graph_warning_title"))
        layout = QVBoxLayout()
        question_label = QLabel()
        question_label.setText(self._("overwrite_graph_warning_text"))
        layout.addWidget(question_label)
        button_box = self.create_button_box()
        button_box.accepted.connect(lambda:self.accept_action(dialog) )
        button_box.rejected.connect(dialog.reject)
        layout.addWidget(button_box)
        dialog.setLayout(layout)
        dialog.show()
        dialog.exec()

    def accept_action(self,dialog:QDialog):
        visuals_screen = global_vars.otl_wizard.main_window.step3_visuals
        current_visuals_html_path = global_vars.current_project.get_current_visuals_html_path()
        visuals_screen.recreate_html(current_visuals_html_path)
        dialog.close()

    def create_button_box(self):
        button_box = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        button_box.setProperty("class", "button-box")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setProperty("class", "primary-button")
        button_box.button(QDialogButtonBox.StandardButton.Ok).setText(self._('continue'))
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setProperty("class", "secondary-button")
        button_box.button(QDialogButtonBox.StandardButton.Cancel).setText(self._("cancel"))
        return button_box

    def validate_documents_again(self, dialog):
        missing_project_files = InsertDataDomain.check_current_project_project_files_existence()
        if len(missing_project_files):
            self.screen.show_missing_project_files_notification_window(missing_project_files)
        else:
            create_task_reraise_exception(self.screen.validate_documents())
        dialog.close()




from abc import abstractmethod

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QWidget, QComboBox, QCheckBox, QRadioButton

from otlmow_gui.Domain.logger.OTLLogger import OTLLogger


class Screen(QWidget):

    # list containing all fields of a screen that need to be reset on calling reset_ui
    # these need to be added manually everytime a developer creates a new field of the following types:
    # QComboBox (dropdownlist)
    # QCheckBox (checkbox)
    resetable_fields_list = []



    def add_resetable_field(self,field ,default_value=None):



        if isinstance(field,QComboBox) and (not default_value or isinstance(default_value,int)):
            if not default_value:
                default_value = 0
        elif isinstance(field, QCheckBox) and (
                        not default_value or isinstance(default_value, bool)):
            if not default_value:
                default_value = False

        elif isinstance(field, QRadioButton) and (
                        not default_value or isinstance(default_value, bool)):
            if not default_value:
                default_value = False
        else:
            field_type= type(field)
            OTLLogger.logger.error(f"Cannot add field type {field_type} to Screen.resetable_fields_list")
            return

        self.resetable_fields_list.append([field, default_value])


    @abstractmethod
    def reset_ui(self, _):
        for field_and_default_value in self.resetable_fields_list:
            field =  field_and_default_value[0]
            default_value =  field_and_default_value[1]

            if isinstance(field, QComboBox) and field.count():
                field.setCurrentIndex(default_value)
            elif isinstance(field, QCheckBox):
                if field.checkState() == Qt.CheckState.Checked:
                    field.setChecked(False)

                if default_value:
                    field.setChecked(True)
            elif isinstance(field, QRadioButton):
                if field.isChecked():
                    field.setChecked(False)

                if default_value:
                    field.setChecked(True)

    def opened(self):
        pass


from enum import Enum


class Language(Enum):
    ENGLISH = 'en'
    DUTCH = 'nl_BE'


class FileState(Enum):
    ERROR = 'error'
    WARNING = 'warning'
    OK = 'ok'


class ReportAction(Enum):
    ATA = 'attribute added'
    ATR = 'attribute removed'
    ATC = 'attribute changed'
    ASS = 'asset added'
    REL = 'relation added'



from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QPushButton


class ButtonWidget(QPushButton):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
